#!/usr/bin/env python

# Author: Remi Flamary <remi.flamary@polytechnique.edu>
#
# License: MIT License

import os
import re
import subprocess
import sys

from setuptools import find_packages, setup
from setuptools.extension import Extension

import numpy
from Cython.Build import cythonize

sys.path.append(os.path.join("ot", "helpers"))
from openmp_helpers import check_openmp_support

# dirty but working
__version__ = re.search(
    r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',  # It excludes inline comment too
    open("ot/__init__.py").read(),
).group(1)
# The beautiful part is, I don't even need to check exceptions here.
# If something messes up, let the build process fail noisy, BEFORE my release!

# thanks PyPI for handling markdown now
ROOT = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(ROOT, "README.md"), encoding="utf-8") as f:
    README = f.read()

# clean cython output is clean is called
if "clean" in sys.argv[1:]:
    for cpp_file in [
        "ot/lp/emd_wrap.cpp",
        "ot/partial/partial_cython.cpp",
    ]:
        if os.path.isfile(cpp_file):
            os.remove(cpp_file)

# add platform dependant optional compilation argument
openmp_supported, flags = check_openmp_support()
compile_args = ["/O2" if sys.platform == "win32" else "-O3"]
link_args = []

if openmp_supported:
    compile_args += flags
    link_args += flags

if sys.platform.startswith("darwin"):
    compile_args.append("-stdlib=libc++")
    sdk_path = subprocess.check_output(["xcrun", "--show-sdk-path"])
    os.environ["CFLAGS"] = '-isysroot "{}"'.format(sdk_path.rstrip().decode("utf-8"))


setup(
    name="POT",
    version=__version__,
    description="Python Optimal Transport Library",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Remi Flamary, Nicolas Courty, Cédric Vincent-Cuaz, POT Contributors",
    author_email="remi.flamary@gmail.com, ncourty@gmail.com, cedvincentcuaz@gmail.com",
    url="https://github.com/PythonOT/POT",
    packages=find_packages(exclude=["benchmarks"]),
    ext_modules=cythonize(
        [
            Extension(
                name="ot.lp.emd_wrap",
                sources=[
                    "ot/lp/emd_wrap.pyx",
                    "ot/lp/EMD_wrapper.cpp",
                ],  # cython/c++ src files
                language="c++",
                include_dirs=[numpy.get_include(), os.path.join(ROOT, "ot/lp")],
                extra_compile_args=compile_args,
                extra_link_args=link_args,
            ),
            Extension(
                name="ot.partial.partial_cython",
                sources=["ot/partial/partial_cython.pyx"],
                include_dirs=[numpy.get_include(), os.path.join(ROOT, "ot/partial")],
                extra_compile_args=compile_args,
                language="c++",
            ),
        ]
    ),
    platforms=["linux", "macosx", "windows"],
    download_url="https://github.com/PythonOT/POT/archive/{}.tar.gz".format(
        __version__
    ),
    license="MIT",
    scripts=[],
    data_files=[],
    install_requires=["numpy>=1.16", "scipy>=1.6"],
    extras_require={
        "backend-numpy": [],  # in requirements.
        "backend-jax": ["jax", "jaxlib"],
        "backend-cupy": [],  # should be installed with conda, not pip
        "backend-tf": ["tensorflow"],
        "backend-torch": ["torch"],
        "cvxopt": ["cvxopt"],  # on it's own to prevent accidental GPL violations
        "dr": ["scikit-learn", "pymanopt", "autograd"],
        "gnn": ["torch", "torch_geometric"],
        "plot": ["matplotlib"],
        "all": [
            "jax",
            "jaxlib",
            "tensorflow",
            "torch",
            "cvxopt",
            "scikit-learn",
            "pymanopt",
            "autograd",
            "torch_geometric",
            "matplotlib",
        ],
    },
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: C++",
        "Programming Language :: C",
        "Programming Language :: Cython",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



# -*- coding: utf-8 -*-
"""
Optimal transport for Gaussian distributions
"""

# Author: Theo Gnassounou <theo.gnassounou@inria.fr>
#         Remi Flamary <remi.flamary@polytehnique.edu>
#
# License: MIT License

import warnings
import numpy as np

from .backend import get_backend
from .utils import dots, is_all_finite, list_to_array, exp_bures


def bures_wasserstein_mapping(ms, mt, Cs, Ct, log=False):
    r"""Return OT linear operator between samples.

    The function estimates the optimal linear operator that aligns the two
    empirical distributions. This is equivalent to estimating the closed
    form mapping between two Gaussian distributions :math:`\mathcal{N}(\mu_s,\Sigma_s)`
    and :math:`\mathcal{N}(\mu_t,\Sigma_t)` as proposed in
    :ref:`[1] <references-OT-mapping-linear>` and discussed in remark 2.29 in
    :ref:`[2] <references-OT-mapping-linear>`.

    The linear operator from source to target :math:`M`

    .. math::
        M(\mathbf{x})= \mathbf{A} \mathbf{x} + \mathbf{b}

    where :

    .. math::
        \mathbf{A} &= \Sigma_s^{-1/2} \left(\Sigma_s^{1/2}\Sigma_t\Sigma_s^{1/2} \right)^{1/2}
        \Sigma_s^{-1/2}

        \mathbf{b} &= \mu_t - \mathbf{A} \mu_s

    Parameters
    ----------
    ms : array-like (d,)
        mean of the source distribution
    mt : array-like (d,)
        mean of the target distribution
    Cs : array-like (d,d)
        covariance of the source distribution
    Ct : array-like (d,d)
        covariance of the target distribution
    log : bool, optional
        record log if True


    Returns
    -------
    A : (d, d) array-like
        Linear operator
    b : (1, d) array-like
        bias
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-OT-mapping-linear:
    References
    ----------
    .. [1] Knott, M. and Smith, C. S. "On the optimal mapping of
        distributions", Journal of Optimization Theory and Applications
        Vol 43, 1984

    .. [2] Peyré, G., & Cuturi, M. (2017). "Computational Optimal
        Transport", 2018.
    """
    ms, mt, Cs, Ct = list_to_array(ms, mt, Cs, Ct)
    nx = get_backend(ms, mt, Cs, Ct)

    Cs12 = nx.sqrtm(Cs)
    Cs12inv = nx.inv(Cs12)

    M0 = nx.sqrtm(dots(Cs12, Ct, Cs12))

    A = dots(Cs12inv, M0, Cs12inv)

    b = mt - nx.dot(ms, A)

    if log:
        log = {}
        log["Cs12"] = Cs12
        log["Cs12inv"] = Cs12inv
        return A, b, log
    else:
        return A, b


def empirical_bures_wasserstein_mapping(
    xs, xt, reg=1e-6, ws=None, wt=None, bias=True, log=False
):
    r"""Return OT linear operator between samples.

    The function estimates the optimal linear operator that aligns the two
    empirical distributions. This is equivalent to estimating the closed
    form mapping between two Gaussian distributions :math:`\mathcal{N}(\mu_s,\Sigma_s)`
    and :math:`\mathcal{N}(\mu_t,\Sigma_t)` as proposed in
    :ref:`[1] <references-OT-mapping-linear>` and discussed in remark 2.29 in
    :ref:`[2] <references-OT-mapping-linear>`.

    The linear operator from source to target :math:`M`

    .. math::
        M(\mathbf{x})= \mathbf{A} \mathbf{x} + \mathbf{b}

    where :

    .. math::
        \mathbf{A} &= \Sigma_s^{-1/2} \left(\Sigma_s^{1/2}\Sigma_t\Sigma_s^{1/2} \right)^{1/2}
        \Sigma_s^{-1/2}

        \mathbf{b} &= \mu_t - \mathbf{A} \mu_s

    Parameters
    ----------
    xs : array-like (ns,d)
        samples in the source domain
    xt : array-like (nt,d)
        samples in the target domain
    reg : float,optional
        regularization added to the diagonals of covariances (>0)
    ws : array-like (ns,1), optional
        weights for the source samples
    wt : array-like (ns,1), optional
        weights for the target samples
    bias: boolean, optional
        estimate bias :math:`\mathbf{b}` else :math:`\mathbf{b} = 0` (default:True)
    log : bool, optional
        record log if True


    Returns
    -------
    A : (d, d) array-like
        Linear operator
    b : (1, d) array-like
        bias
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-OT-mapping-linear:
    References
    ----------
    .. [1] Knott, M. and Smith, C. S. "On the optimal mapping of
        distributions", Journal of Optimization Theory and Applications
        Vol 43, 1984

    .. [2] Peyré, G., & Cuturi, M. (2017). "Computational Optimal
        Transport", 2018.
    """
    xs, xt = list_to_array(xs, xt)
    nx = get_backend(xs, xt)
    is_input_finite = is_all_finite(xs, xt)

    d = xs.shape[1]

    if ws is None:
        ws = nx.ones((xs.shape[0], 1), type_as=xs) / xs.shape[0]

    if wt is None:
        wt = nx.ones((xt.shape[0], 1), type_as=xt) / xt.shape[0]

    if bias:
        mxs = nx.dot(ws.T, xs) / nx.sum(ws)
        mxt = nx.dot(wt.T, xt) / nx.sum(wt)

        xs = xs - mxs
        xt = xt - mxt
    else:
        mxs = nx.zeros((1, d), type_as=xs)
        mxt = nx.zeros((1, d), type_as=xs)

    Cs = nx.dot((xs * ws).T, xs) / nx.sum(ws) + reg * nx.eye(d, type_as=xs)
    Ct = nx.dot((xt * wt).T, xt) / nx.sum(wt) + reg * nx.eye(d, type_as=xt)

    if log:
        A, b, log = bures_wasserstein_mapping(mxs, mxt, Cs, Ct, log=log)
    else:
        A, b = bures_wasserstein_mapping(mxs, mxt, Cs, Ct)

    if is_input_finite and not is_all_finite(A, b):
        warnings.warn(
            "Numerical errors were encountered in ot.gaussian.empirical_bures_wasserstein_mapping. "
            "Consider increasing the regularization parameter `reg`."
        )

    if log:
        log["Cs"] = Cs
        log["Ct"] = Ct
        return A, b, log
    else:
        return A, b


def bures_distance(Cs, Ct, paired=False, log=False, nx=None):
    r"""Return Bures distance.

    The function computes the Bures distance between :math:`\mu_s=\mathcal{N}(0,\Sigma_s)` and :math:`\mu_t=\mathcal{N}(0,\Sigma_t)`,
    given by (see e.g. Remark 2.31 :ref:`[15] <references-bures-wasserstein-distance>`):

    .. math::
        \mathbf{B}(\Sigma_s, \Sigma_t)^{2} = \text{Tr}\left(\Sigma_s + \Sigma_t - 2 \sqrt{\Sigma_s^{1/2}\Sigma_t\Sigma_s^{1/2}} \right)

    Parameters
    ----------
    Cs : array-like (d,d) or (n,d,d)
        covariance of the source distribution
    Ct : array-like (d,d) or (m,d,d)
        covariance of the target distribution
    paired: bool, optional
        if True and n==m, return the paired distances and crossed distance otherwise
    log : bool, optional
        record log if True
    nx : module, optional
        The numerical backend module to use. If not provided, the backend will
        be fetched from the input matrices `Cs, Ct`.

    Returns
    -------
    W : float if Cs and Cd of shape (d,d), array-like (n,m) if Cs of shape (n,d,d) and Ct of shape (m,d,d), array-like (n,) if Cs and Ct of shape (n, d, d) and paired is True
        Bures Wasserstein distance
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-bures-wasserstein-distance:
    References
    ----------
    .. [15] Peyré, G., & Cuturi, M. (2019). Computational optimal transport: With applications to data science.
        Foundations and Trends® in Machine Learning, 11(5-6), 355-607.
    """
    Cs, Ct = list_to_array(Cs, Ct)

    if nx is None:
        nx = get_backend(Cs, Ct)

    assert Cs.shape[-1] == Ct.shape[-1], "All Gaussian must have the same dimension"

    Cs12 = nx.sqrtm(Cs)

    if len(Cs.shape) == 2 and len(Ct.shape) == 2:
        # Return float
        bw2 = nx.trace(Cs + Ct - 2 * nx.sqrtm(dots(Cs12, Ct, Cs12)))
    else:
        assert (
            len(Cs.shape) == 3 and len(Ct.shape) == 3
        ), "Both Cs and Ct should be batched"
        if paired and len(Cs) == len(Ct):
            # Return shape (n,)
            M = nx.einsum("nij, njk, nkl -> nil", Cs12, Ct, Cs12)
            bw2 = nx.trace(Cs + Ct - 2 * nx.sqrtm(M))
        else:
            # Return shape (n,m)
            M = nx.einsum("nij, mjk, nkl -> nmil", Cs12, Ct, Cs12)
            bw2 = nx.trace(Cs[:, None] + Ct[None] - 2 * nx.sqrtm(M))

    W = nx.sqrt(nx.maximum(bw2, 0))

    if log:
        log = {}
        log["Cs12"] = Cs12
        return W, log
    else:
        return W


def bures_wasserstein_distance(ms, mt, Cs, Ct, paired=False, log=False):
    r"""Return Bures Wasserstein distance between samples.

    The function computes the Bures-Wasserstein distance between :math:`\mu_s=\mathcal{N}(m_s,\Sigma_s)` and :math:`\mu_t=\mathcal{N}(m_t,\Sigma_t)`,
    as discussed in remark 2.31 :ref:`[15] <references-bures-wasserstein-distance>`.

    .. math::
        \mathcal{W}(\mu_s, \mu_t)_2^2= \left\lVert \mathbf{m}_s - \mathbf{m}_t \right\rVert^2 + \mathcal{B}(\Sigma_s, \Sigma_t)^{2}

    where :

    .. math::
        \mathbf{B}(\Sigma_s, \Sigma_t)^{2} = \text{Tr}\left(\Sigma_s + \Sigma_t - 2 \sqrt{\Sigma_s^{1/2}\Sigma_t\Sigma_s^{1/2}} \right)

    Parameters
    ----------
    ms : array-like (d,) or (n,d)
        mean of the source distribution
    mt : array-like (d,) or (m,d)
        mean of the target distribution
    Cs : array-like (d,d) or (n,d,d)
        covariance of the source distribution
    Ct : array-like (d,d) or (m,d,d)
        covariance of the target distribution
    paired: bool, optional
        if True and n==m, return the paired distances and crossed distance otherwise
    log : bool, optional
        record log if True

    Returns
    -------
    W : float if ms and md of shape (d,), array-like (n,m) if ms of shape (n,d) and mt of shape (m,d), array-like (n,) if ms and mt of shape (n,d) and paired is True
        Bures Wasserstein distance
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-bures-wasserstein-distance:
    References
    ----------

    .. [15] Peyré, G., & Cuturi, M. (2019). Computational optimal transport: With applications to data science.
        Foundations and Trends® in Machine Learning, 11(5-6), 355-607.
    """
    ms, mt, Cs, Ct = list_to_array(ms, mt, Cs, Ct)
    nx = get_backend(ms, mt, Cs, Ct)

    assert (
        ms.shape[0] == Cs.shape[0]
    ), "Source Gaussians has different amount of components"

    assert (
        mt.shape[0] == Ct.shape[0]
    ), "Target Gaussians has different amount of components"

    assert (
        ms.shape[-1] == mt.shape[-1] == Cs.shape[-1] == Ct.shape[-1]
    ), "All Gaussian must have the same dimension"

    if log:
        bw, log_dict = bures_distance(Cs, Ct, paired=paired, log=log, nx=nx)
        Cs12 = log_dict["Cs12"]
    else:
        bw = bures_distance(Cs, Ct, paired=paired, nx=nx)

    if len(ms.shape) == 1 and len(mt.shape) == 1:
        # Return float
        squared_dist_m = nx.norm(ms - mt) ** 2
    else:
        assert (
            len(ms.shape) == 2 and len(mt.shape) == 2
        ), "Both ms and mt should be batched"
        if paired and len(ms.shape) == len(mt.shape):
            # Return shape (n,)
            squared_dist_m = nx.norm(ms - mt, axis=-1) ** 2
        else:
            # Return shape (n,m)
            squared_dist_m = nx.norm(ms[:, None] - mt[None], axis=-1) ** 2

    W = nx.sqrt(nx.maximum(squared_dist_m + bw**2, 0))

    if log:
        log = {}
        log["Cs12"] = Cs12
        return W, log
    else:
        return W


def empirical_bures_wasserstein_distance(
    xs, xt, reg=1e-6, ws=None, wt=None, bias=True, log=False
):
    r"""Return Bures Wasserstein distance from mean and covariance of distribution.

    The function estimates the Bures-Wasserstein distance between two
    empirical distributions source :math:`\mu_s` and target :math:`\mu_t`,
    discussed in remark 2.31 :ref:`[1] <references-bures-wasserstein-distance>`.

    The Bures Wasserstein distance between source and target distribution :math:`\mathcal{W}`

    .. math::
        \mathcal{W}(\mu_s, \mu_t)_2^2= \left\lVert \mathbf{m}_s - \mathbf{m}_t \right\rVert^2 + \mathcal{B}(\Sigma_s, \Sigma_t)^{2}

    where :

    .. math::
        \mathbf{B}(\Sigma_s, \Sigma_t)^{2} = \text{Tr}\left(\Sigma_s + \Sigma_t - 2 \sqrt{\Sigma_s^{1/2}\Sigma_t\Sigma_s^{1/2}} \right)

    Parameters
    ----------
    xs : array-like (ns,d)
        samples in the source domain
    xt : array-like (nt,d)
        samples in the target domain
    reg : float,optional
        regularization added to the diagonals of covariances (>0)
    ws : array-like (ns), optional
        weights for the source samples
    wt : array-like (ns), optional
        weights for the target samples
    bias: boolean, optional
        estimate bias :math:`\mathbf{b}` else :math:`\mathbf{b} = 0` (default:True)
    log : bool, optional
        record log if True


    Returns
    -------
    W : float
        Bures Wasserstein distance
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-bures-wasserstein-distance:
    References
    ----------

    .. [1] Peyré, G., & Cuturi, M. (2017). "Computational Optimal
        Transport", 2018.
    """
    xs, xt = list_to_array(xs, xt)
    nx = get_backend(xs, xt)

    d = xs.shape[1]

    if ws is None:
        ws = nx.ones((xs.shape[0], 1), type_as=xs) / xs.shape[0]

    if wt is None:
        wt = nx.ones((xt.shape[0], 1), type_as=xt) / xt.shape[0]

    if bias:
        mxs = nx.dot(ws.T, xs) / nx.sum(ws)
        mxt = nx.dot(wt.T, xt) / nx.sum(wt)

        xs = xs - mxs
        xt = xt - mxt
    else:
        mxs = nx.zeros((1, d), type_as=xs)
        mxt = nx.zeros((1, d), type_as=xs)

    Cs = nx.dot((xs * ws).T, xs) / nx.sum(ws) + reg * nx.eye(d, type_as=xs)
    Ct = nx.dot((xt * wt).T, xt) / nx.sum(wt) + reg * nx.eye(d, type_as=xt)

    if log:
        W, log = bures_wasserstein_distance(mxs[0], mxt[0], Cs, Ct, log=log)
        log["Cs"] = Cs
        log["Ct"] = Ct
        return W, log
    else:
        W = bures_wasserstein_distance(mxs[0], mxt[0], Cs, Ct)
        return W


def bures_barycenter_fixpoint(
    C, weights=None, num_iter=1000, eps=1e-7, log=False, nx=None
):
    r"""Return the (Bures-)Wasserstein barycenter between centered Gaussian distributions.

    The function estimates the (Bures)-Wasserstein barycenter between centered Gaussian distributions :math:`\big(\mathcal{N}(0,\Sigma_i)\big)_{i=1}^n`
    :ref:`[16] <references-OT-bures-barycenter-fixed-point>` by solving

    .. math::
        \Sigma_b = \mathrm{argmin}_{\Sigma \in S_d^{++}(\mathbb{R})}\ \sum_{i=1}^n w_i W_2^2\big(\mathcal{N}(0,\Sigma), \mathcal{N}(0, \Sigma_i)\big).

    The barycenter still follows a Gaussian distribution :math:`\mathcal{N}(0,\Sigma_b)`
    where :math:`\Sigma_b` is solution of the following fixed-point algorithm:

    .. math::
        \Sigma_b = \sum_{i=1}^n w_i \left(\Sigma_b^{1/2}\Sigma_i^{1/2}\Sigma_b^{1/2}\right)^{1/2}.

    Parameters
    ----------
    C : array-like (k,d,d)
        covariance of k distributions
    weights : array-like (k), optional
        weights for each distribution
    method : str
        method used for the solver, either 'fixed_point' or 'gradient_descent'
    num_iter : int, optional
        number of iteration for the fixed point algorithm
    eps : float, optional
        tolerance for the fixed point algorithm
    log : bool, optional
        record log if True
    nx : module, optional
        The numerical backend module to use. If not provided, the backend will
        be fetched from the input matrices `C`.

    Returns
    -------
    Cb : (d, d) array-like
        covariance of the barycenter
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-OT-bures-barycenter-fixed-point:
    References
    ----------
    .. [16] M. Agueh and G. Carlier, "Barycenters in the Wasserstein space",
        SIAM Journal on Mathematical Analysis, vol. 43, no. 2, pp. 904-924,
        2011.
    """
    if nx is None:
        nx = get_backend(
            *C,
        )

    if weights is None:
        weights = nx.ones(C.shape[0], type_as=C[0]) / C.shape[0]

    # Init the covariance barycenter
    Cb = nx.mean(C * weights[:, None, None], axis=0)

    for it in range(num_iter):
        # fixed point update
        Cb12 = nx.sqrtm(Cb)

        Cnew = nx.sqrtm(Cb12 @ C @ Cb12)
        Cnew *= weights[:, None, None]
        Cnew = nx.sum(Cnew, axis=0)

        # check convergence
        diff = nx.norm(Cb - Cnew)
        if diff <= eps:
            break
        Cb = Cnew

    if diff > eps:
        print("Dit not converge.")

    if log:
        log = {}
        log["num_iter"] = it
        log["final_diff"] = diff
        return Cb, log
    else:
        return Cb


def bures_barycenter_gradient_descent(
    C,
    weights=None,
    num_iter=1000,
    eps=1e-7,
    log=False,
    step_size=1,
    batch_size=None,
    averaged=False,
    nx=None,
):
    r"""Return the (Bures-)Wasserstein barycenter between centered Gaussian distributions.

    The function estimates the (Bures)-Wasserstein barycenter between centered Gaussian distributions :math:`\big(\mathcal{N}(0,\Sigma_i)\big)_{i=1}^n`
    by using a gradient descent in the Wasserstein space :ref:`[74, 75] <references-OT-bures-barycenter-gradient_descent>`
    on the objective

    .. math::
        \mathcal{L}(\Sigma) = \sum_{i=1}^n w_i W_2^2\big(\mathcal{N}(0,\Sigma), \mathcal{N}(0,\Sigma_i)\big).

    Parameters
    ----------
    C : array-like (k,d,d)
        covariance of k distributions
    weights : array-like (k), optional
        weights for each distribution
    method : str
        method used for the solver, either 'fixed_point' or 'gradient_descent'
    num_iter : int, optional
        number of iteration for the fixed point algorithm
    eps : float, optional
        tolerance for the fixed point algorithm
    log : bool, optional
        record log if True
    step_size : float, optional
        step size for the gradient descent, 1 by default
    batch_size : int, optional
        batch size if use a stochastic gradient descent
    averaged : bool, optional
        if True, use the averaged procedure of :ref:`[74] <references-OT-bures-barycenter-gradient_descent>`
    nx : module, optional
        The numerical backend module to use. If not provided, the backend will
        be fetched from the input matrices `C`.

    Returns
    -------
    Cb : (d, d) array-like
        covariance of the barycenter
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-OT-bures-barycenter-gradient_descent:
    References
    ----------
    .. [74] Chewi, S., Maunu, T., Rigollet, P., & Stromme, A. J. (2020).
        Gradient descent algorithms for Bures-Wasserstein barycenters.
        In Conference on Learning Theory (pp. 1276-1304). PMLR.

    .. [75] Altschuler, J., Chewi, S., Gerber, P. R., & Stromme, A. (2021).
        Averaging on the Bures-Wasserstein manifold: dimension-free convergence
        of gradient descent. Advances in Neural Information Processing Systems, 34, 22132-22145.
    """
    if nx is None:
        nx = get_backend(
            *C,
        )

    n = C.shape[0]

    if weights is None:
        weights = nx.ones(C.shape[0], type_as=C[0]) / n

    # Init the covariance barycenter
    Cb = nx.mean(C * weights[:, None, None], axis=0)
    Id = nx.eye(C.shape[-1], type_as=Cb)

    L_diff = []

    Cb_averaged = nx.copy(Cb)

    for it in range(num_iter):
        Cb12 = nx.sqrtm(Cb)
        Cb12_ = nx.inv(Cb12)

        if batch_size is not None and batch_size < n:  # if stochastic gradient descent
            if batch_size <= 0:
                raise ValueError(
                    "batch_size must be an integer between 0 and {}".format(n)
                )
            inds = np.random.choice(
                n, batch_size, replace=True, p=nx._to_numpy(weights)
            )
            M = nx.sqrtm(nx.einsum("ij,njk,kl -> nil", Cb12, C[inds], Cb12))
            ot_maps = nx.einsum("ij,njk,kl -> nil", Cb12_, M, Cb12_)
            grad_bw = Id - nx.mean(ot_maps, axis=0)

            # step size from [74] (page 15)
            step_size = 2 / (0.7 * (it + 2 / 0.7 + 1))
        else:  # gradient descent
            M = nx.sqrtm(nx.einsum("ij,njk,kl -> nil", Cb12, C, Cb12))
            ot_maps = nx.einsum("ij,njk,kl -> nil", Cb12_, M, Cb12_)
            grad_bw = Id - nx.sum(ot_maps * weights[:, None, None], axis=0)

        Cnew = exp_bures(Cb, -step_size * grad_bw, nx=nx)

        if averaged:
            # ot map between Cb_averaged and Cnew
            Cb_averaged12 = nx.sqrtm(Cb_averaged)
            Cb_averaged12inv = nx.inv(Cb_averaged12)
            M = nx.sqrtm(nx.einsum("ij,jk,kl->il", Cb_averaged12, Cnew, Cb_averaged12))
            ot_map = nx.einsum("ij,jk,kl->il", Cb_averaged12inv, M, Cb_averaged12inv)
            map = Id * step_size / (step_size + 1) + ot_map / (step_size + 1)
            Cb_averaged = nx.einsum("ij,jk,kl->il", map, Cb_averaged, map)

        # check convergence
        L_diff.append(nx.norm(Cb - Cnew))

        # Criteria to stop
        if np.mean(L_diff[-100:]) <= eps:
            break

        Cb = Cnew

    if averaged:
        Cb = Cb_averaged

    if log:
        dict_log = {}
        dict_log["num_iter"] = it
        dict_log["final_diff"] = L_diff[-1]
        return Cb, dict_log
    else:
        return Cb


def bures_wasserstein_barycenter(
    m,
    C,
    weights=None,
    method="fixed_point",
    num_iter=1000,
    eps=1e-7,
    log=False,
    step_size=1,
    batch_size=None,
):
    r"""Return the (Bures-)Wasserstein barycenter between Gaussian distributions.

    The function estimates the (Bures)-Wasserstein barycenter between Gaussian distributions :math:`\big(\mathcal{N}(\mu_i,\Sigma_i)\big)_{i=1}^n`
    :ref:`[16, 74, 75] <references-OT-bures_wasserstein-barycenter>` by solving

    .. math::
        (\mu_b, \Sigma_b) = \mathrm{argmin}_{\mu,\Sigma}\ \sum_{i=1}^n w_i W_2^2\big(\mathcal{N}(\mu,\Sigma), \mathcal{N}(\mu_i, \Sigma_i)\big)

    The barycenter still follows a Gaussian distribution :math:`\mathcal{N}(\mu_b,\Sigma_b)`
    where:

    .. math::
        \mu_b = \sum_{i=1}^n w_i \mu_i,

    and the barycentric covariance is the solution of the following fixed-point algorithm:

    .. math::
        \Sigma_b = \sum_{i=1}^n w_i \left(\Sigma_b^{1/2}\Sigma_i^{1/2}\Sigma_b^{1/2}\right)^{1/2}

    We propose two solvers: one based on solving the previous fixed-point problem [16]. Another based on
    gradient descent in the Bures-Wasserstein space [74,75].

    Parameters
    ----------
    m : array-like (k,d)
        mean of k distributions
    C : array-like (k,d,d)
        covariance of k distributions
    weights : array-like (k), optional
        weights for each distribution
    method : str
        method used for the solver, either 'fixed_point', 'gradient_descent', 'stochastic_gradient_descent' or
        'averaged_stochastic_gradient_descent'
    num_iter : int, optional
        number of iteration for the fixed point algorithm
    eps : float, optional
        tolerance for the fixed point algorithm
    log : bool, optional
        record log if True
    step_size : float, optional
        step size for the gradient descent, 1 by default
    batch_size : int, optional
        batch size if use a stochastic gradient descent. If not None, use method='gradient_descent'


    Returns
    -------
    mb : (d,) array-like
        mean of the barycenter
    Cb : (d, d) array-like
        covariance of the barycenter
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-OT-bures_wasserstein-barycenter:
    References
    ----------
    .. [16] M. Agueh and G. Carlier, "Barycenters in the Wasserstein space",
        SIAM Journal on Mathematical Analysis, vol. 43, no. 2, pp. 904-924,
        2011.

    .. [74] Chewi, S., Maunu, T., Rigollet, P., & Stromme, A. J. (2020).
        Gradient descent algorithms for Bures-Wasserstein barycenters.
        In Conference on Learning Theory (pp. 1276-1304). PMLR.

    .. [75] Altschuler, J., Chewi, S., Gerber, P. R., & Stromme, A. (2021).
        Averaging on the Bures-Wasserstein manifold: dimension-free convergence
        of gradient descent. Advances in Neural Information Processing Systems, 34, 22132-22145.
    """
    nx = get_backend(
        *m,
    )

    if weights is None:
        weights = nx.ones(C.shape[0], type_as=C[0]) / C.shape[0]

    # Compute the mean barycenter
    mb = nx.sum(m * weights[:, None], axis=0)

    if method == "gradient_descent":
        out = bures_barycenter_gradient_descent(
            C,
            weights=weights,
            num_iter=num_iter,
            eps=eps,
            log=log,
            step_size=step_size,
            nx=nx,
        )
    elif method == "stochastic_gradient_descent":
        out = bures_barycenter_gradient_descent(
            C,
            weights=weights,
            num_iter=num_iter,
            eps=eps,
            log=log,
            batch_size=1 if batch_size is None else batch_size,
            nx=nx,
        )
    elif method == "averaged_stochastic_gradient_descent":
        out = bures_barycenter_gradient_descent(
            C,
            weights=weights,
            num_iter=num_iter,
            eps=eps,
            log=log,
            batch_size=1 if batch_size is None else batch_size,
            averaged=True,
            nx=nx,
        )
    elif method == "fixed_point":
        out = bures_barycenter_fixpoint(
            C, weights=weights, num_iter=num_iter, eps=eps, log=log, nx=nx
        )
    else:
        raise ValueError("Unknown method '%s'." % method)

    if log:
        Cb, log = out
        return mb, Cb, log
    else:
        Cb = out
        return mb, Cb


def empirical_bures_wasserstein_barycenter(
    X, reg=1e-6, weights=None, num_iter=1000, eps=1e-7, w=None, bias=True, log=False
):
    r"""Return OT linear operator between samples.

    The function estimates the optimal barycenter of the
    empirical distributions. This is equivalent to resolving the fixed point
    algorithm for multiple Gaussian distributions :math:`\left\{\mathcal{N}(\mu,\Sigma)\right\}_{i=1}^n`
    :ref:`[1] <references-OT-mapping-linear-barycenter>`.

    The barycenter still following a Gaussian distribution :math:`\mathcal{N}(\mu_b,\Sigma_b)`
    where :

    .. math::
        \mu_b = \sum_{i=1}^n w_i \mu_i

    And the barycentric covariance is the solution of the following fixed-point algorithm:

    .. math::
        \Sigma_b = \sum_{i=1}^n w_i \left(\Sigma_b^{1/2}\Sigma_i^{1/2}\Sigma_b^{1/2}\right)^{1/2}


    Parameters
    ----------
    X : list of array-like (n,d)
        samples in each distribution
    reg : float,optional
        regularization added to the diagonals of covariances (>0)
    weights : array-like (n,), optional
        weights for each distribution
    num_iter : int, optional
        number of iteration for the fixed point algorithm
    eps : float, optional
        tolerance for the fixed point algorithm
    w : list of array-like (n,), optional
        weights for each sample in each distribution
    bias: boolean, optional
        estimate bias :math:`\mathbf{b}` else :math:`\mathbf{b} = 0` (default:True)
    log : bool, optional
        record log if True


    Returns
    -------
    mb : (d,) array-like
        mean of the barycenter
    Cb : (d, d) array-like
        covariance of the barycenter
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-OT-mapping-linear-barycenter:
    References
    ----------
    .. [1] M. Agueh and G. Carlier, "Barycenters in the Wasserstein space",
        SIAM Journal on Mathematical Analysis, vol. 43, no. 2, pp. 904-924,
        2011.
    """
    X = list_to_array(*X)
    nx = get_backend(*X)

    k = len(X)
    d = [X[i].shape[1] for i in range(k)]

    if w is None:
        w = [
            nx.ones((X[i].shape[0], 1), type_as=X[i]) / X[i].shape[0] for i in range(k)
        ]

    if bias:
        m = [nx.dot(w[i].T, X[i]) / nx.sum(w[i]) for i in range(k)]
        X = [X[i] - m[i] for i in range(k)]
    else:
        m = [nx.zeros((1, d[i]), type_as=X[i]) for i in range(k)]

    C = [
        nx.dot((X[i] * w[i]).T, X[i]) / nx.sum(w[i]) + reg * nx.eye(d[i], type_as=X[i])
        for i in range(k)
    ]
    m = nx.stack(m, axis=0)[:, 0]
    C = nx.stack(C, axis=0)

    if log:
        mb, Cb, log = bures_wasserstein_barycenter(
            m, C, weights=weights, num_iter=num_iter, eps=eps, log=log
        )

        return mb, Cb, log
    else:
        mb, Cb = bures_wasserstein_barycenter(
            m, C, weights=weights, num_iter=num_iter, eps=eps, log=log
        )
        return mb, Cb


def gaussian_gromov_wasserstein_distance(Cov_s, Cov_t, log=False):
    r"""Return the Gaussian Gromov-Wasserstein value from [57].

    This function return the closed form value of the Gaussian Gromov-Wasserstein
    distance between two Gaussian distributions
    :math:`\mathcal{N}(\mu_s,\Sigma_s)` and :math:`\mathcal{N}(\mu_t,\Sigma_t)`
    when the OT plan is assumed to be also Gaussian. See [57] Theorem 4.1 for
    more details.

    Parameters
    ----------
    Cov_s : array-like (ds,ds)
        covariance of the source distribution
    Cov_t : array-like (dt,dt)
        covariance of the target distribution


    Returns
    -------
    G : float
        Gaussian Gromov-Wasserstein distance


    .. _references-gaussien_gromov_wasserstein_distance:
    References
    ----------
    .. [57] Delon, J., Desolneux, A., & Salmona, A. (2022). Gromov–Wasserstein distances between Gaussian distributions.
        Journal of Applied Probability, 59(4), 1178-1198.
    """

    nx = get_backend(Cov_s, Cov_t)

    # ensure that Cov_s is the largest covariance matrix
    # that is m >= n
    if Cov_s.shape[0] < Cov_t.shape[0]:
        Cov_s, Cov_t = Cov_t, Cov_s

    n = Cov_t.shape[0]

    # compte and sort eigenvalues decerasingly
    d_s = nx.flip(nx.sort(nx.eigh(Cov_s)[0]))
    d_t = nx.flip(nx.sort(nx.eigh(Cov_t)[0]))

    # compute the gaussien Gromov-Wasserstein distance
    res = (
        4 * (nx.sum(d_s) - nx.sum(d_t)) ** 2
        + 8 * nx.sum((d_s[:n] - d_t) ** 2)
        + 8 * nx.sum((d_s[n:]) ** 2)
    )
    if log:
        log = {}
        log["d_s"] = d_s
        log["d_t"] = d_t
        return nx.sqrt(res), log
    else:
        return nx.sqrt(res)


def empirical_gaussian_gromov_wasserstein_distance(xs, xt, ws=None, wt=None, log=False):
    r"""Return Gaussian Gromov-Wasserstein distance between samples.

    The function estimates the Gaussian Gromov-Wasserstein distance between two
    Gaussien distributions source :math:`\mu_s` and target :math:`\mu_t`, whose
    parameters are estimated from the provided samples :math:`\mathcal{X}_s` and
    :math:`\mathcal{X}_t`. See [57] Theorem 4.1 for more details.

    Parameters
    ----------
    xs : array-like (ns,d)
        samples in the source domain
    xt : array-like (nt,d)
        samples in the target domain
    ws : array-like (ns,1), optional
        weights for the source samples
    wt : array-like (ns,1), optional
        weights for the target samples
    log : bool, optional
        record log if True


    Returns
    -------
    G : float
        Gaussian Gromov-Wasserstein distance


    .. _references-gaussien_gromov_wasserstein:
    References
    ----------
    .. [57] Delon, J., Desolneux, A., & Salmona, A. (2022).
        Gromov–Wasserstein distances between Gaussian distributions.
        Journal of Applied Probability, 59(4), 1178-1198.
    """
    xs, xt = list_to_array(xs, xt)
    nx = get_backend(xs, xt)

    if ws is None:
        ws = nx.ones((xs.shape[0], 1), type_as=xs) / xs.shape[0]

    if wt is None:
        wt = nx.ones((xt.shape[0], 1), type_as=xt) / xt.shape[0]

    mxs = nx.dot(ws.T, xs) / nx.sum(ws)
    mxt = nx.dot(wt.T, xt) / nx.sum(wt)

    xs = xs - mxs
    xt = xt - mxt

    Cs = nx.dot((xs * ws).T, xs) / nx.sum(ws)
    Ct = nx.dot((xt * wt).T, xt) / nx.sum(wt)

    if log:
        G, log = gaussian_gromov_wasserstein_distance(Cs, Ct, log=log)
        log["Cov_s"] = Cs
        log["Cov_t"] = Ct
        return G, log
    else:
        G = gaussian_gromov_wasserstein_distance(Cs, Ct)
        return G


def gaussian_gromov_wasserstein_mapping(
    mu_s, mu_t, Cov_s, Cov_t, sign_eigs=None, log=False
):
    r"""Return the Gaussian Gromov-Wasserstein mapping from [57].

    This function return the closed form value of the Gaussian
    Gromov-Wasserstein mapping between two Gaussian distributions
    :math:`\mathcal{N}(\mu_s,\Sigma_s)` and :math:`\mathcal{N}(\mu_t,\Sigma_t)`
    when the OT plan is assumed to be also Gaussian. See [57] Theorem 4.1 for
    more details.

    Parameters
    ----------
    mu_s : array-like (ds,)
        mean of the source distribution
    mu_t : array-like (dt,)
        mean of the target distribution
    Cov_s : array-like (ds,ds)
        covariance of the source distribution
    Cov_t : array-like (dt,dt)
        covariance of the target distribution
    log : bool, optional
        record log if True


    Returns
    -------
    A : (dt, ds) array-like
        Linear operator
    b : (1, dt) array-like
        bias


    .. _references-gaussien_gromov_wasserstein_mapping:
    References
    ----------
    .. [57] Delon, J., Desolneux, A., & Salmona, A. (2022).
        Gromov–Wasserstein distances between Gaussian distributions.
        Journal of Applied Probability, 59(4), 1178-1198.
    """

    nx = get_backend(mu_s, mu_t, Cov_s, Cov_t)

    n = Cov_t.shape[0]
    m = Cov_s.shape[0]

    # compte and sort eigenvalues/eigenvectors decreasingly
    d_s, U_s = nx.eigh(Cov_s)
    id_s = nx.flip(nx.argsort(d_s))
    d_s, U_s = d_s[id_s], U_s[:, id_s]

    d_t, U_t = nx.eigh(Cov_t)
    id_t = nx.flip(nx.argsort(d_t))
    d_t, U_t = d_t[id_t], U_t[:, id_t]

    if sign_eigs is None:
        sign_eigs = nx.ones(min(m, n), type_as=mu_s)

    if m >= n:
        A = nx.concatenate(
            (
                nx.diag(sign_eigs * nx.sqrt(d_t) / nx.sqrt(d_s[:n])),
                nx.zeros((n, m - n), type_as=mu_s),
            ),
            axis=1,
        ).T
    else:
        A = nx.concatenate(
            (
                nx.diag(sign_eigs * nx.sqrt(d_t[:m]) / nx.sqrt(d_s)),
                nx.zeros((n - m, m), type_as=mu_s),
            ),
            axis=0,
        ).T

    A = nx.dot(nx.dot(U_s, A), U_t.T)

    # compute the gaussien Gromov-Wasserstein dis
    b = mu_t - nx.dot(mu_s, A)

    if log:
        log = {}
        log["d_s"] = d_s
        log["d_t"] = d_t
        log["U_s"] = U_s
        log["U_t"] = U_t
        return A, b, log
    else:
        return A, b


def empirical_gaussian_gromov_wasserstein_mapping(
    xs, xt, ws=None, wt=None, sign_eigs=None, log=False
):
    r"""Return Gaussian Gromov-Wasserstein mapping between samples.

    The function estimates the Gaussian Gromov-Wasserstein mapping between two
    Gaussian distributions source :math:`\mu_s` and target :math:`\mu_t`, whose
    parameters are estimated from the provided samples :math:`\mathcal{X}_s` and
    :math:`\mathcal{X}_t`. See [57] Theorem 4.1 for more details.


    Parameters
    ----------
    xs : array-like (ns,ds)
        samples in the source domain
    xt : array-like (nt,dt)
        samples in the target domain
    ws : array-like (ns,1), optional
        weights for the source samples
    wt : array-like (ns,1), optional
        weights for the target samples
    sign_eigs : array-like (min(ds,dt),) or string, optional
        sign of the eigenvalues of the mapping matrix, by default all signs will
        be positive. If 'skewness' is provided, the sign of the eigenvalues is
        selected as the product of the sign of the skewness of the projected data.
    log : bool, optional
        record log if True


    Returns
    -------
    A : (dt, ds) array-like
        Linear operator
    b : (1, dt) array-like
        bias


    .. _references-empirical_gaussian_gromov_wasserstein_mapping:
    References
    ----------
    .. [57] Delon, J., Desolneux, A., & Salmona, A. (2022).
        Gromov–Wasserstein distances between Gaussian distributions.
        Journal of Applied Probability, 59(4), 1178-1198.
    """

    xs, xt = list_to_array(xs, xt)

    nx = get_backend(xs, xt)

    m = xs.shape[1]
    n = xt.shape[1]

    if ws is None:
        ws = nx.ones((xs.shape[0], 1), type_as=xs) / xs.shape[0]

    if wt is None:
        wt = nx.ones((xt.shape[0], 1), type_as=xt) / xt.shape[0]

    # estimate mean and covariance
    mu_s = nx.dot(ws.T, xs) / nx.sum(ws)
    mu_t = nx.dot(wt.T, xt) / nx.sum(wt)

    xs = xs - mu_s
    xt = xt - mu_t

    Cov_s = nx.dot((xs * ws).T, xs) / nx.sum(ws)
    Cov_t = nx.dot((xt * wt).T, xt) / nx.sum(wt)

    # compute and sort eigenvalues/eigenvectors decreasingly
    d_s, U_s = nx.eigh(Cov_s)
    id_s = nx.flip(nx.argsort(d_s))
    d_s, U_s = d_s[id_s], U_s[:, id_s]

    d_t, U_t = nx.eigh(Cov_t)
    id_t = nx.flip(nx.argsort(d_t))
    d_t, U_t = d_t[id_t], U_t[:, id_t]

    # select the sign of the eigenvalues
    if sign_eigs is None:
        sign_eigs = nx.ones(min(m, n), type_as=mu_s)
    elif sign_eigs == "skewness":
        size = min(m, n)
        skew_s = nx.sum((nx.dot(xs, U_s[:, :size])) ** 3 * ws, axis=0)
        skew_t = nx.sum((nx.dot(xt, U_t[:, :size])) ** 3 * wt, axis=0)
        sign_eigs = nx.sign(skew_t * skew_s)

    if m >= n:
        A = nx.concatenate(
            (
                nx.diag(sign_eigs * nx.sqrt(d_t) / nx.sqrt(d_s[:n])),
                nx.zeros((n, m - n), type_as=mu_s),
            ),
            axis=1,
        ).T
    else:
        A = nx.concatenate(
            (
                nx.diag(sign_eigs * nx.sqrt(d_t[:m]) / nx.sqrt(d_s)),
                nx.zeros((n - m, m), type_as=mu_s),
            ),
            axis=0,
        ).T

    A = nx.dot(nx.dot(U_s, A), U_t.T)

    # compute the gaussien Gromov-Wasserstein dis
    b = mu_t - nx.dot(mu_s, A)

    if log:
        log = {}
        log["d_s"] = d_s
        log["d_t"] = d_t
        log["U_s"] = U_s
        log["U_t"] = U_t
        log["Cov_s"] = Cov_s
        log["Cov_t"] = Cov_t
        return A, b, log
    else:
        return A, b



# -*- coding: utf-8 -*-
"""
Regularized Unbalanced OT solvers
"""

# Author: Hicham Janati <hicham.janati@inria.fr>
#         Laetitia Chapel <laetitia.chapel@univ-ubs.fr>
#         Quang Huy Tran <quang-huy.tran@univ-ubs.fr>
#
# License: MIT License

from ..backend import get_backend
from ..utils import list_to_array, get_parameter_pair


def mm_unbalanced(
    a,
    b,
    M,
    reg_m,
    c=None,
    reg=0,
    div="kl",
    G0=None,
    numItermax=1000,
    stopThr=1e-15,
    verbose=False,
    log=False,
):
    r"""
    Solve the unbalanced optimal transport problem and return the OT plan.
    The function solves the following optimization problem:

    .. math::
        W = \arg \min_\gamma \quad \langle \gamma, \mathbf{M} \rangle_F +
        \mathrm{reg_{m1}} \cdot \mathrm{div}(\gamma \mathbf{1}, \mathbf{a}) +
        \mathrm{reg_{m2}} \cdot \mathrm{div}(\gamma^T \mathbf{1}, \mathbf{b}) +
        \mathrm{reg} \cdot \mathrm{div}(\gamma, \mathbf{c})

        s.t.
             \gamma \geq 0

    where:

    - :math:`\mathbf{M}` is the (`dim_a`, `dim_b`) metric cost matrix
    - :math:`\mathbf{a}` and :math:`\mathbf{b}` are source and target
      unbalanced distributions
    - :math:`\mathbf{c}` is a reference distribution for the regularization
    - div is a divergence, either Kullback-Leibler or half-squared :math:`\ell_2` divergence

    The algorithm used for solving the problem is a maximization-
    minimization algorithm as proposed in :ref:`[41] <references-regpath>`

    Parameters
    ----------
    a : array-like (dim_a,)
        Unnormalized histogram of dimension `dim_a`
        If `a` is an empty list or array ([]),
        then `a` is set to uniform distribution.
    b : array-like (dim_b,)
        Unnormalized histogram of dimension `dim_b`
        If `b` is an empty list or array ([]),
        then `b` is set to uniform distribution.
    M : array-like (dim_a, dim_b)
        loss matrix
    reg_m: float or indexable object of length 1 or 2
        Marginal relaxation term: nonnegative but cannot be infinity.
        If :math:`\mathrm{reg_{m}}` is a scalar or an indexable object of length 1,
        then the same :math:`\mathrm{reg_{m}}` is applied to both marginal relaxations.
        If :math:`\mathrm{reg_{m}}` is an array,
        it must have the same backend as input arrays `(a, b, M)`.
    reg : float, optional (default = 0)
        Regularization term >= 0.
        By default, solve the unregularized problem
    c : array-like (dim_a, dim_b), optional (default = None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = \mathbf{a} \mathbf{b}^T`.
    div: string, optional
        Divergence to quantify the difference between the marginals.
        Can take two values: 'kl' (Kullback-Leibler) or 'l2' (half-squared)
    G0: array-like (dim_a, dim_b)
        Initialization of the transport matrix
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (> 0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record log if True
    Returns
    -------
    gamma : (dim_a, dim_b) array-like
            Optimal transportation matrix for the given parameters
    log : dict
            log dictionary returned only if `log` is `True`

    Examples
    --------
    >>> import ot
    >>> import numpy as np
    >>> a=[.5, .5]
    >>> b=[.5, .5]
    >>> M=[[1., 36.],[9., 4.]]
    >>> np.round(ot.unbalanced.mm_unbalanced(a, b, M, 5, div='kl'), 2)
    array([[0.45, 0.  ],
           [0.  , 0.34]])
    >>> np.round(ot.unbalanced.mm_unbalanced(a, b, M, 5, div='l2'), 2)
    array([[0.4, 0. ],
           [0. , 0.1]])


    .. _references-regpath:
    References
    ----------
    .. [41] Chapel, L., Flamary, R., Wu, H., Févotte, C., and Gasso, G. (2021).
        Unbalanced optimal transport through non-negative penalized
        linear regression. NeurIPS.

    See Also
    --------
    ot.lp.emd : Unregularized OT
    ot.unbalanced.sinkhorn_unbalanced : Entropic regularized OT
    """

    M, a, b = list_to_array(M, a, b)
    nx = get_backend(M, a, b)

    dim_a, dim_b = M.shape

    if len(a) == 0:
        a = nx.ones(dim_a, type_as=M) / dim_a
    if len(b) == 0:
        b = nx.ones(dim_b, type_as=M) / dim_b

    G = a[:, None] * b[None, :] if G0 is None else G0
    if reg > 0:  # regularized case
        c = a[:, None] * b[None, :] if c is None else c
    else:  # unregularized case
        c = 0

    reg_m1, reg_m2 = get_parameter_pair(reg_m)

    if log:
        log = {"err": [], "G": []}

    div = div.lower()
    if div == "kl":
        sum_r = reg + reg_m1 + reg_m2
        r1, r2, r = reg_m1 / sum_r, reg_m2 / sum_r, reg / sum_r
        K = (a[:, None] ** r1) * (b[None, :] ** r2) * (c**r) * nx.exp(-M / sum_r)
    elif div == "l2":
        K = (reg_m1 * a[:, None]) + (reg_m2 * b[None, :]) + reg * c - M
        K = nx.maximum(K, nx.zeros((dim_a, dim_b), type_as=M))
    else:
        raise ValueError("Unknown div = {}. Must be either 'kl' or 'l2'".format(div))

    for i in range(numItermax):
        Gprev = G

        if div == "kl":
            Gd = (nx.sum(G, 1, keepdims=True) ** r1) * (
                nx.sum(G, 0, keepdims=True) ** r2
            ) + 1e-16
            G = K * G ** (r1 + r2) / Gd
        elif div == "l2":
            Gd = (
                reg_m1 * nx.sum(G, 1, keepdims=True)
                + reg_m2 * nx.sum(G, 0, keepdims=True)
                + reg * G
                + 1e-16
            )
            G = K * G / Gd

        err = nx.sqrt(nx.sum((G - Gprev) ** 2))
        if log:
            log["err"].append(err)
            log["G"].append(G)
        if verbose:
            print("{:5d}|{:8e}|".format(i, err))
        if err < stopThr:
            break

    if log:
        linear_cost = nx.sum(G * M)
        log["cost"] = linear_cost

        m1, m2 = nx.sum(G, 1), nx.sum(G, 0)
        if div == "kl":
            cost = (
                linear_cost
                + reg_m1 * nx.kl_div(m1, a, mass=True)
                + reg_m2 * nx.kl_div(m2, b, mass=True)
            )
            if reg > 0:
                cost = cost + reg * nx.kl_div(G, c, mass=True)
        else:
            cost = (
                linear_cost
                + reg_m1 * 0.5 * nx.sum((m1 - a) ** 2)
                + reg_m2 * 0.5 * nx.sum((m2 - b) ** 2)
            )
            if reg > 0:
                cost = cost + reg * 0.5 * nx.sum((G - c) ** 2)

        log["total_cost"] = cost

        return G, log
    else:
        return G


def mm_unbalanced2(
    a,
    b,
    M,
    reg_m,
    c=None,
    reg=0,
    div="kl",
    G0=None,
    returnCost="linear",
    numItermax=1000,
    stopThr=1e-15,
    verbose=False,
    log=False,
):
    r"""
    Solve the unbalanced optimal transport problem and return the OT cost.
    The function solves the following optimization problem:

    .. math::
        \min_\gamma \quad \langle \gamma, \mathbf{M} \rangle_F +
        \mathrm{reg_{m1}} \cdot \mathrm{div}(\gamma \mathbf{1}, \mathbf{a}) +
        \mathrm{reg_{m2}} \cdot \mathrm{div}(\gamma^T \mathbf{1}, \mathbf{b}) +
        \mathrm{reg} \cdot \mathrm{div}(\gamma, \mathbf{c})

        s.t.
             \gamma \geq 0

    where:

    - :math:`\mathbf{M}` is the (`dim_a`, `dim_b`) metric cost matrix
    - :math:`\mathbf{a}` and :math:`\mathbf{b}` are source and target
      unbalanced distributions
    - :math:`\mathbf{c}` is a reference distribution for the regularization
    - :math:`\mathrm{div}` is a divergence, either Kullback-Leibler or half-squared :math:`\ell_2` divergence

    The algorithm used for solving the problem is a maximization-
    minimization algorithm as proposed in :ref:`[41] <references-regpath>`

    Parameters
    ----------
    a : array-like (dim_a,)
        Unnormalized histogram of dimension `dim_a`
        If `a` is an empty list or array ([]),
        then `a` is set to uniform distribution.
    b : array-like (dim_b,)
        Unnormalized histogram of dimension `dim_b`
        If `b` is an empty list or array ([]),
        then `b` is set to uniform distribution.
    M : array-like (dim_a, dim_b)
        loss matrix
    reg_m: float or indexable object of length 1 or 2
        Marginal relaxation term: nonnegative but cannot be infinity.
        If :math:`\mathrm{reg_{m}}` is a scalar or an indexable object of length 1,
        then the same :math:`\mathrm{reg_{m}}` is applied to both marginal relaxations.
        If :math:`\mathrm{reg_{m}}` is an array,
        it must have the same backend as input arrays `(a, b, M)`.
    reg : float, optional (default = 0)
        Entropy regularization term >= 0.
        By default, solve the unregularized problem
    c : array-like (dim_a, dim_b), optional (default = None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = mathbf{a} mathbf{b}^T`.
    div: string, optional
        Divergence to quantify the difference between the marginals.
        Can take two values: 'kl' (Kullback-Leibler) or 'l2' (half-squared)
    G0: array-like (dim_a, dim_b)
        Initialization of the transport matrix
    returnCost: string, optional (default = "linear")
        If `returnCost` = "linear", then return the linear part of the unbalanced OT loss.
        If `returnCost` = "total", then return the total unbalanced OT loss.
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (> 0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record log if True

    Returns
    -------
    ot_cost : array-like
        the OT cost between :math:`\mathbf{a}` and :math:`\mathbf{b}`
    log : dict
        log dictionary returned only if `log` is `True`

    Examples
    --------
    >>> import ot
    >>> import numpy as np
    >>> a=[.5, .5]
    >>> b=[.5, .5]
    >>> M=[[1., 36.],[9., 4.]]
    >>> np.round(ot.unbalanced.mm_unbalanced2(a, b, M, 5, div='l2'), 2)
    0.8
    >>> np.round(ot.unbalanced.mm_unbalanced2(a, b, M, 5, div='kl'), 2)
    1.79

    References
    ----------
    .. [41] Chapel, L., Flamary, R., Wu, H., Févotte, C., and Gasso, G. (2021).
        Unbalanced optimal transport through non-negative penalized
        linear regression. NeurIPS.
    See Also
    --------
    ot.lp.emd2 : Unregularized OT loss
    ot.unbalanced.sinkhorn_unbalanced2 : Entropic regularized OT loss
    """

    _, log_mm = mm_unbalanced(
        a,
        b,
        M,
        reg_m,
        c=c,
        reg=reg,
        div=div,
        G0=G0,
        numItermax=numItermax,
        stopThr=stopThr,
        verbose=verbose,
        log=True,
    )

    if returnCost == "linear":
        cost = log_mm["cost"]
    elif returnCost == "total":
        cost = log_mm["total_cost"]
    else:
        raise ValueError("Unknown returnCost = {}".format(returnCost))

    if log:
        return cost, log_mm
    else:
        return cost



# -*- coding: utf-8 -*-
"""
Regularized Unbalanced OT solvers
"""

# Author: Hicham Janati <hicham.janati@inria.fr>
#         Laetitia Chapel <laetitia.chapel@univ-ubs.fr>
#         Quang Huy Tran <quang-huy.tran@univ-ubs.fr>
#
# License: MIT License

import numpy as np
from scipy.optimize import minimize, Bounds

from ..backend import get_backend
from ..utils import list_to_array, get_parameter_pair, fun_to_numpy


def _get_loss_unbalanced(a, b, c, M, reg, reg_m1, reg_m2, reg_div="kl", regm_div="kl"):
    r"""
    Return loss function for the L-BFGS-B solver

    .. note:: This function will be fed into scipy.optimize, so all input arrays must be Numpy arrays.

    Parameters
    ----------
    a : array-like (dim_a,)
        Unnormalized histogram of dimension `dim_a`
    b : array-like (dim_b,)
        Unnormalized histogram of dimension `dim_b`
    M : array-like (dim_a, dim_b)
        loss matrix
    reg: float
        regularization term >=0
    c : array-like (dim_a, dim_b), optional (default = None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = \mathbf{a} \mathbf{b}^T`.
    reg_m1: float
        Marginal relaxation term with respect to the first marginal:
        nonnegative (including 0) but cannot be infinity.
    reg_m2: float
        Marginal relaxation term with respect to the second marginal:
        nonnegative (including 0) but cannot be infinity.
    reg_div: string, optional
        Divergence used for regularization.
        Can take three values: 'entropy' (negative entropy), or
        'kl' (Kullback-Leibler) or 'l2' (half-squared) or a tuple
        of two callable functions returning the reg term and its derivative.
        Note that the callable functions should be able to handle Numpy arrays
        and not tensors from the backend
    regm_div: string, optional
        Divergence to quantify the difference between the marginals.
        Can take three values: 'kl' (Kullback-Leibler) or 'l2' (half-squared) or 'tv' (Total Variation)

    Returns
    -------
    Loss function (scipy.optimize compatible) for regularized unbalanced OT
    """

    m, n = M.shape
    nx_numpy = get_backend(M, a, b)

    def reg_l2(G):
        return np.sum((G - c) ** 2) / 2

    def grad_l2(G):
        return G - c

    def reg_kl(G):
        return nx_numpy.kl_div(G, c, mass=True)

    def grad_kl(G):
        return np.log(G / c + 1e-16)

    def reg_entropy(G):
        return np.sum(G * np.log(G + 1e-16)) - np.sum(G)

    def grad_entropy(G):
        return np.log(G + 1e-16)

    if reg_div == "kl":
        reg_fun = reg_kl
        grad_reg_fun = grad_kl
    elif reg_div == "entropy":
        reg_fun = reg_entropy
        grad_reg_fun = grad_entropy
    elif isinstance(reg_div, tuple):
        reg_fun = reg_div[0]
        grad_reg_fun = reg_div[1]
    else:
        reg_fun = reg_l2
        grad_reg_fun = grad_l2

    def marg_l2(G):
        return reg_m1 * 0.5 * np.sum((G.sum(1) - a) ** 2) + reg_m2 * 0.5 * np.sum(
            (G.sum(0) - b) ** 2
        )

    def grad_marg_l2(G):
        return reg_m1 * np.outer((G.sum(1) - a), np.ones(n)) + reg_m2 * np.outer(
            np.ones(m), (G.sum(0) - b)
        )

    def marg_kl(G):
        return reg_m1 * nx_numpy.kl_div(
            G.sum(1), a, mass=True
        ) + reg_m2 * nx_numpy.kl_div(G.sum(0), b, mass=True)

    def grad_marg_kl(G):
        return reg_m1 * np.outer(
            np.log(G.sum(1) / a + 1e-16), np.ones(n)
        ) + reg_m2 * np.outer(np.ones(m), np.log(G.sum(0) / b + 1e-16))

    def marg_tv(G):
        return reg_m1 * np.sum(np.abs(G.sum(1) - a)) + reg_m2 * np.sum(
            np.abs(G.sum(0) - b)
        )

    def grad_marg_tv(G):
        return reg_m1 * np.outer(np.sign(G.sum(1) - a), np.ones(n)) + reg_m2 * np.outer(
            np.ones(m), np.sign(G.sum(0) - b)
        )

    if regm_div == "kl":
        regm_fun = marg_kl
        grad_regm_fun = grad_marg_kl
    elif regm_div == "tv":
        regm_fun = marg_tv
        grad_regm_fun = grad_marg_tv
    else:
        regm_fun = marg_l2
        grad_regm_fun = grad_marg_l2

    def _func(G):
        G = G.reshape((m, n))

        # compute loss
        val = np.sum(G * M) + regm_fun(G)
        if reg > 0:
            val = val + reg * reg_fun(G)
        # compute gradient
        grad = M + grad_regm_fun(G)
        if reg > 0:
            grad = grad + reg * grad_reg_fun(G)

        return val, grad.ravel()

    return _func


def lbfgsb_unbalanced(
    a,
    b,
    M,
    reg,
    reg_m,
    c=None,
    reg_div="kl",
    regm_div="kl",
    G0=None,
    numItermax=1000,
    stopThr=1e-15,
    method="L-BFGS-B",
    verbose=False,
    log=False,
):
    r"""
    Solve the unbalanced optimal transport problem and return the OT plan using L-BFGS-B algorithm.
    The function solves the following optimization problem:

    .. math::
        W = \arg \min_\gamma \quad \langle \gamma, \mathbf{M} \rangle_F +
        \mathrm{reg} \mathrm{div}(\gamma, \mathbf{c}) +
        \mathrm{reg_{m1}} \cdot \mathrm{div_m}(\gamma \mathbf{1}, \mathbf{a}) +
        \mathrm{reg_{m2}} \cdot \mathrm{div_m}(\gamma^T \mathbf{1}, \mathbf{b})

        s.t.
             \gamma \geq 0

    where:

    - :math:`\mathbf{M}` is the (`dim_a`, `dim_b`) metric cost matrix
    - :math:`\mathbf{a}` and :math:`\mathbf{b}` are source and target unbalanced distributions
    - :math:`\mathbf{c}` is a reference distribution for the regularization
    - :math:`\mathrm{div_m}` is a divergence, either Kullback-Leibler divergence,
    or half-squared :math:`\ell_2` divergence, or Total variation
    - :math:`\mathrm{div}` is a divergence, either Kullback-Leibler divergence,
    or half-squared :math:`\ell_2` divergence

    .. note:: This function is backend-compatible and will work on arrays
        from all compatible backends. First, it converts all arrays into Numpy arrays,
        then uses the L-BFGS-B algorithm from scipy.optimize to solve the optimization problem.

    Parameters
    ----------
    a : array-like (dim_a,)
        Unnormalized histogram of dimension `dim_a`
        If `a` is an empty list or array ([]),
        then `a` is set to uniform distribution.
    b : array-like (dim_b,)
        Unnormalized histogram of dimension `dim_b`
        If `b` is an empty list or array ([]),
        then `b` is set to uniform distribution.
    M : array-like (dim_a, dim_b)
        loss matrix
    reg: float
        regularization term >=0
    reg_m: float or indexable object of length 1 or 2
        Marginal relaxation term: nonnegative (including 0) but cannot be infinity.
        If :math:`\mathrm{reg_{m}}` is a scalar or an indexable object of length 1,
        then the same :math:`\mathrm{reg_{m}}` is applied to both marginal relaxations.
        If :math:`\mathrm{reg_{m}}` is an array, it must be a Numpy array.
    c : array-like (dim_a, dim_b), optional (default = None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = \mathbf{a} \mathbf{b}^T`.
    reg_div: string or pair of callable functions, optional (default = 'kl')
        Divergence used for regularization.
        Can take three values: 'entropy' (negative entropy), or
        'kl' (Kullback-Leibler) or 'l2' (half-squared) or a tuple
        of two callable functions returning the reg term and its derivative.
        Note that the callable functions should be able to handle Numpy arrays
        and not tensors from the backend, otherwise functions will be converted to Numpy
        leading to a computational overhead.
    regm_div: string, optional (default = 'kl')
        Divergence to quantify the difference between the marginals.
        Can take three values: 'kl' (Kullback-Leibler) or 'l2' (half-squared) or 'tv' (Total Variation)
    G0: array-like (dim_a, dim_b), optional (default = None)
        Initialization of the transport matrix. None corresponds to uniform product.
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (> 0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record log if True

    Returns
    -------
    gamma : (dim_a, dim_b) array-like
            Optimal transportation matrix for the given parameters
    log : dict
        log dictionary returned only if `log` is `True`

    Examples
    --------
    >>> import ot
    >>> import numpy as np
    >>> a=[.5, .5]
    >>> b=[.5, .5]
    >>> M=[[1., 36.],[9., 4.]]
    >>> np.round(ot.unbalanced.lbfgsb_unbalanced(a, b, M, reg=0, reg_m=5, reg_div='kl', regm_div='kl'), 2)
    array([[0.45, 0.  ],
           [0.  , 0.34]])
    >>> np.round(ot.unbalanced.lbfgsb_unbalanced(a, b, M, reg=0, reg_m=5, reg_div='l2', regm_div='l2'), 2)
    array([[0.4, 0. ],
           [0. , 0.1]])

    References
    ----------
    .. [41] Chapel, L., Flamary, R., Wu, H., Févotte, C., and Gasso, G. (2021).
        Unbalanced optimal transport through non-negative penalized
        linear regression. NeurIPS.
    See Also
    --------
    ot.lp.emd2 : Unregularized OT loss
    ot.unbalanced.sinkhorn_unbalanced2 : Entropic regularized OT loss
    """

    # test settings
    regm_div = regm_div.lower()
    if regm_div not in ["kl", "l2", "tv"]:
        raise ValueError(
            "Unknown regm_div = {}. Must be either 'kl', 'l2' or 'tv'".format(regm_div)
        )

    if isinstance(reg_div, str):
        reg_div = reg_div.lower()
        if reg_div not in ["entropy", "kl", "l2"]:
            raise ValueError(
                "Unknown reg_div = {}. Must be either 'entropy', 'kl' or 'l2', or a tuple".format(
                    reg_div
                )
            )

    # convert all inputs to numpy arrays
    reg_m1, reg_m2 = get_parameter_pair(reg_m)

    M, a, b = list_to_array(M, a, b)
    nx = get_backend(M, a, b, G0)
    M0 = M

    dim_a, dim_b = M.shape

    if len(a) == 0:
        a = nx.ones(dim_a, type_as=M) / dim_a
    if len(b) == 0:
        b = nx.ones(dim_b, type_as=M) / dim_b

    # convert to numpy
    if nx.__name__ == "numpy":  # remaining parameters which can be arrays
        reg_m1, reg_m2, reg = nx.to_numpy(reg_m1, reg_m2, reg)
    else:
        a, b, M, reg_m1, reg_m2, reg = nx.to_numpy(a, b, M, reg_m1, reg_m2, reg)

    G0 = a[:, None] * b[None, :] if G0 is None else nx.to_numpy(G0)
    c = a[:, None] * b[None, :] if c is None else nx.to_numpy(c)

    # potentially convert the callable function to handle numpy arrays
    if isinstance(reg_div, tuple):
        f0, df0 = reg_div
        f = fun_to_numpy(f0, G0, nx, warn=True)
        df = fun_to_numpy(df0, G0, nx, warn=True)

        reg_div = (f, df)

    _func = _get_loss_unbalanced(a, b, c, M, reg, reg_m1, reg_m2, reg_div, regm_div)

    res = minimize(
        _func,
        G0.ravel(),
        method=method,
        jac=True,
        bounds=Bounds(0, np.inf),
        tol=stopThr,
        options=dict(maxiter=numItermax, disp=verbose),
    )

    G = nx.from_numpy(res.x.reshape(M.shape), type_as=M0)

    if log:
        log = {"cost": nx.sum(G * M), "res": res}
        log["total_cost"] = nx.from_numpy(res.fun, type_as=M0)
        return G, log
    else:
        return G


def lbfgsb_unbalanced2(
    a,
    b,
    M,
    reg,
    reg_m,
    c=None,
    reg_div="kl",
    regm_div="kl",
    G0=None,
    returnCost="linear",
    numItermax=1000,
    stopThr=1e-15,
    method="L-BFGS-B",
    verbose=False,
    log=False,
):
    r"""
    Solve the unbalanced optimal transport problem and return the OT cost using L-BFGS-B.
    The function solves the following optimization problem:

    .. math::
        \min_\gamma \quad \langle \gamma, \mathbf{M} \rangle_F +
        \mathrm{reg} \mathrm{div}(\gamma, \mathbf{c}) +
        \mathrm{reg_{m1}} \cdot \mathrm{div_m}(\gamma \mathbf{1}, \mathbf{a}) +
        \mathrm{reg_{m2}} \cdot \mathrm{div_m}(\gamma^T \mathbf{1}, \mathbf{b})

        s.t.
             \gamma \geq 0

    where:

    - :math:`\mathbf{M}` is the (`dim_a`, `dim_b`) metric cost matrix
    - :math:`\mathbf{a}` and :math:`\mathbf{b}` are source and target unbalanced distributions
    - :math:`\mathbf{c}` is a reference distribution for the regularization
    - :math:`\mathrm{div_m}` is a divergence, either Kullback-Leibler divergence,
    or half-squared :math:`\ell_2` divergence, or Total variation
    - :math:`\mathrm{div}` is a divergence, either Kullback-Leibler divergence,
    or half-squared :math:`\ell_2` divergence

    .. note:: This function is backend-compatible and will work on arrays
        from all compatible backends. First, it converts all arrays into Numpy arrays,
        then uses the L-BFGS-B algorithm from scipy.optimize to solve the optimization problem.

    Parameters
    ----------
    a : array-like (dim_a,)
        Unnormalized histogram of dimension `dim_a`
        If `a` is an empty list or array ([]),
        then `a` is set to uniform distribution.
    b : array-like (dim_b,)
        Unnormalized histogram of dimension `dim_b`
        If `b` is an empty list or array ([]),
        then `b` is set to uniform distribution.
    M : array-like (dim_a, dim_b)
        loss matrix
    reg: float
        regularization term >=0
    reg_m: float or indexable object of length 1 or 2
        Marginal relaxation term: nonnegative (including 0) but cannot be infinity.
        If :math:`\mathrm{reg_{m}}` is a scalar or an indexable object of length 1,
        then the same :math:`\mathrm{reg_{m}}` is applied to both marginal relaxations.
        If :math:`\mathrm{reg_{m}}` is an array, it must be a Numpy array.
    c : array-like (dim_a, dim_b), optional (default = None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = \mathbf{a} \mathbf{b}^T`.
    reg_div: string or pair of callable functions, optional (default = 'kl')
        Divergence used for regularization.
        Can take three values: 'entropy' (negative entropy), or
        'kl' (Kullback-Leibler) or 'l2' (half-squared) or a tuple
        of two callable functions returning the reg term and its derivative.
        Note that the callable functions should be able to handle Numpy arrays
        and not tensors from the backend, otherwise functions will be converted to Numpy
        leading to a computational overhead.
    regm_div: string, optional (default = 'kl')
        Divergence to quantify the difference between the marginals.
        Can take three values: 'kl' (Kullback-Leibler) or 'l2' (half-squared) or 'tv' (Total Variation)
    G0: array-like (dim_a, dim_b), optional (default = None)
        Initialization of the transport matrix. None corresponds to uniform product.
    returnCost: string, optional (default = "linear")
        If `returnCost` = "linear", then return the linear part of the unbalanced OT loss.
        If `returnCost` = "total", then return the total unbalanced OT loss.
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (> 0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record log if True

    Returns
    -------
    ot_cost : array-like
        the OT cost between :math:`\mathbf{a}` and :math:`\mathbf{b}`
    log : dict
        log dictionary returned only if `log` is `True`

    Examples
    --------
    >>> import ot
    >>> import numpy as np
    >>> a=[.5, .5]
    >>> b=[.5, .5]
    >>> M=[[1., 36.],[9., 4.]]
    >>> np.round(ot.unbalanced.lbfgsb_unbalanced2(a, b, M, reg=0, reg_m=5, reg_div='kl', regm_div='kl'), 2)
    1.79
    >>> np.round(ot.unbalanced.lbfgsb_unbalanced2(a, b, M, reg=0, reg_m=5, reg_div='l2', regm_div='l2'), 2)
    0.8

    References
    ----------
    .. [41] Chapel, L., Flamary, R., Wu, H., Févotte, C., and Gasso, G. (2021).
        Unbalanced optimal transport through non-negative penalized
        linear regression. NeurIPS.
    See Also
    --------
    ot.lp.emd2 : Unregularized OT loss
    ot.unbalanced.sinkhorn_unbalanced2 : Entropic regularized OT loss
    """

    _, log_lbfgs = lbfgsb_unbalanced(
        a=a,
        b=b,
        M=M,
        reg=reg,
        reg_m=reg_m,
        c=c,
        reg_div=reg_div,
        regm_div=regm_div,
        G0=G0,
        numItermax=numItermax,
        stopThr=stopThr,
        method=method,
        verbose=verbose,
        log=True,
    )

    if returnCost == "linear":
        cost = log_lbfgs["cost"]
    elif returnCost == "total":
        cost = log_lbfgs["total_cost"]
    else:
        raise ValueError("Unknown returnCost = {}".format(returnCost))

    if log:
        return cost, log_lbfgs
    else:
        return cost



"""Test for low rank sinkhorn solvers"""

# Author: Laurène DAVID <laurene.david@ip-paris.fr>
#         Titouan Vayer <titouan.vayer@inria.fr>
#
# License: MIT License

import ot
import numpy as np
import pytest
from ot.lowrank import sklearn_import  # check sklearn installation


def test_nystroem_kernel_approx():
    # test nystroem kernel approx in easy regime (nb anchors = nb points)
    n = 30
    d = 3
    Xs = np.random.randn(n, d)
    Xt = np.random.randn(n, d) + 2
    sigma = 2.0
    K = np.exp(-ot.dist(Xs, Xt) / (2 * sigma**2))
    U, V = ot.lowrank.kernel_nystroem(Xs, Xt, anchors=60, sigma=sigma, random_state=42)

    np.testing.assert_allclose(K, U @ V.T, atol=1e-7)


@pytest.mark.parametrize("log, warn", [[False, False], [True, True]])
def test_nystroem_sinkhorn(log, warn):
    # test Nystrom approximation for Sinkhorn (ot plan)
    offset = 2
    n_samples_per_blob = 50
    random_state = 42
    std = 0.1
    np.random.seed(random_state)

    centers = np.array(
        [
            [-offset, -offset],  # Class 0 - blob 1
            [-offset, offset],  # Class 0 - blob 2
            [offset, -offset],  # Class 1 - blob 1
            [offset, offset],  # Class 1 - blob 2
        ]
    )

    X_list = []
    y_list = []

    for i, center in enumerate(centers):
        blob_points = np.random.randn(n_samples_per_blob, 2) * std + center
        label = 0 if i < 2 else 1
        X_list.append(blob_points)
        y_list.append(np.full(n_samples_per_blob, label))

    X = np.vstack(X_list)
    y = np.concatenate(y_list)
    Xs = X[y == 0]
    Xt = X[y == 1]

    reg = 5.0
    anchors = 5

    res = ot.bregman.empirical_sinkhorn_nystroem(
        Xs,
        Xt,
        anchors=anchors,
        reg=reg,
        numItermax=3000,
        verbose=True,
        random_state=random_state,
        log=log,
        warn=warn,
    )
    if log:
        G_nys, log_ = res
    else:
        G_nys = res

    G_sinkh = ot.bregman.empirical_sinkhorn(
        Xs, Xt, reg=reg, numIterMax=3000, verbose=True
    )

    a = ot.unif(Xs.shape[0])
    b = ot.unif(Xt.shape[0])

    np.testing.assert_allclose(G_sinkh, G_nys[:], atol=1e-04)
    np.testing.assert_allclose(a, G_nys[:].sum(1), atol=1e-05)
    np.testing.assert_allclose(b, G_nys[:].sum(0), atol=1e-05)


@pytest.mark.parametrize("log", [False, True])
def test_nystroem_sinkhorn2(log):
    # test Nystrom approximation for Sinkhorn (loss)
    offset = 2
    n_samples_per_blob = 50
    random_state = 42
    std = 0.1
    np.random.seed(random_state)

    centers = np.array(
        [
            [-offset, -offset],  # Class 0 - blob 1
            [-offset, offset],  # Class 0 - blob 2
            [offset, -offset],  # Class 1 - blob 1
            [offset, offset],  # Class 1 - blob 2
        ]
    )

    X_list = []
    y_list = []

    for i, center in enumerate(centers):
        blob_points = np.random.randn(n_samples_per_blob, 2) * std + center
        label = 0 if i < 2 else 1
        X_list.append(blob_points)
        y_list.append(np.full(n_samples_per_blob, label))

    X = np.vstack(X_list)
    y = np.concatenate(y_list)
    Xs = X[y == 0]
    Xt = X[y == 1]

    reg = 5.0
    anchors = 15

    res = ot.bregman.empirical_sinkhorn_nystroem2(
        Xs,
        Xt,
        anchors=anchors,
        reg=reg,
        numItermax=3000,
        verbose=True,
        random_state=random_state,
        log=log,
    )

    if log:
        loss1, log_ = res
    else:
        loss1 = res

    loss2 = ot.bregman.empirical_sinkhorn2(
        Xs, Xt, reg=reg, numIterMax=3000, verbose=True
    )

    np.testing.assert_allclose(loss1, loss2, atol=1e-07, rtol=1e-3)

    with pytest.raises(ValueError, match="anchors must"):
        res = ot.bregman.empirical_sinkhorn_nystroem2(
            Xs,
            Xt,
            anchors=1,
            reg=reg,
            numItermax=3000,
            verbose=True,
            random_state=random_state,
            log=log,
        )


def test_compute_lr_sqeuclidean_matrix():
    # test computation of low rank cost matrices M1 and M2
    n = 100
    X_s = np.reshape(1.0 * np.arange(2 * n), (n, 2))
    X_t = np.reshape(1.0 * np.arange(2 * n), (n, 2))

    M1, M2 = ot.lowrank.compute_lr_sqeuclidean_matrix(X_s, X_t, rescale_cost=False)
    M = ot.dist(X_s, X_t, metric="sqeuclidean")  # original cost matrix

    np.testing.assert_allclose(np.dot(M1, M2.T), M, atol=1e-05)


def test_lowrank_sinkhorn():
    # test low rank sinkhorn
    n = 100
    a = ot.unif(n)
    b = ot.unif(n)

    X_s = np.reshape(1.0 * np.arange(n), (n, 1))
    X_t = np.reshape(1.0 * np.arange(n), (n, 1))

    Q, R, g, log = ot.lowrank.lowrank_sinkhorn(
        X_s, X_t, a, b, reg=0.1, log=True, rescale_cost=False
    )
    P = log["lazy_plan"][:]
    value_linear = log["value_linear"]

    # check constraints for P
    np.testing.assert_allclose(a, P.sum(1), atol=1e-05)
    np.testing.assert_allclose(b, P.sum(0), atol=1e-05)

    # check if lazy_plan is equal to the fully computed plan
    P_true = np.dot(Q, np.dot(np.diag(1 / g), R.T))
    np.testing.assert_allclose(P, P_true, atol=1e-05)

    # check if value_linear is correct with its original formula
    M = ot.dist(X_s, X_t, metric="sqeuclidean")
    value_linear_true = np.sum(M * P_true)
    np.testing.assert_allclose(value_linear, value_linear_true, atol=1e-05)

    # check warn parameter when Dykstra algorithm doesn't converge
    with pytest.warns(UserWarning):
        ot.lowrank.lowrank_sinkhorn(X_s, X_t, a, b, reg=0.1, stopThr=0, numItermax=1)


@pytest.mark.parametrize(("init"), ("random", "deterministic", "kmeans"))
def test_lowrank_sinkhorn_init(init):
    # test lowrank inits
    n = 100
    a = ot.unif(n)
    b = ot.unif(n)

    X_s = np.reshape(1.0 * np.arange(n), (n, 1))
    X_t = np.reshape(1.0 * np.arange(n), (n, 1))

    # test ImportError if init="kmeans" and sklearn not imported
    if init in ["random", "deterministic"] or (
        (init == "kmeans") and (sklearn_import is True)
    ):
        Q, R, g, log = ot.lowrank.lowrank_sinkhorn(
            X_s, X_t, a, b, reg=0.1, init=init, log=True
        )
        P = log["lazy_plan"][:]

        # check constraints for P
        np.testing.assert_allclose(a, P.sum(1), atol=1e-05)
        np.testing.assert_allclose(b, P.sum(0), atol=1e-05)

    else:
        with pytest.raises(ImportError):
            Q, R, g = ot.lowrank.lowrank_sinkhorn(X_s, X_t, a, b, reg=0.1, init=init)


@pytest.mark.parametrize(("alpha, rank"), ((0.8, 2), (0.5, 3), (0.2, 6)))
def test_lowrank_sinkhorn_alpha_error(alpha, rank):
    # Test warning for value of alpha
    n = 100
    a = ot.unif(n)
    b = ot.unif(n)

    X_s = np.reshape(1.0 * np.arange(n), (n, 1))
    X_t = np.reshape(1.0 * np.arange(0, n), (n, 1))

    with pytest.raises(ValueError):
        ot.lowrank.lowrank_sinkhorn(
            X_s, X_t, a, b, reg=0.1, rank=rank, alpha=alpha, warn=False
        )


@pytest.mark.parametrize(("gamma_init"), ("rescale", "theory"))
def test_lowrank_sinkhorn_gamma_init(gamma_init):
    # Test lr sinkhorn with different init strategies
    n = 100
    a = ot.unif(n)
    b = ot.unif(n)

    X_s = np.reshape(1.0 * np.arange(n), (n, 1))
    X_t = np.reshape(1.0 * np.arange(n), (n, 1))

    Q, R, g, log = ot.lowrank.lowrank_sinkhorn(
        X_s, X_t, a, b, reg=0.1, gamma_init=gamma_init, log=True
    )
    P = log["lazy_plan"][:]

    # check constraints for P
    np.testing.assert_allclose(a, P.sum(1), atol=1e-05)
    np.testing.assert_allclose(b, P.sum(0), atol=1e-05)


@pytest.skip_backend("tf")
def test_lowrank_sinkhorn_backends(nx):
    # Test low rank sinkhorn for different backends
    n = 100
    a = ot.unif(n)
    b = ot.unif(n)

    X_s = np.reshape(1.0 * np.arange(n), (n, 1))
    X_t = np.reshape(1.0 * np.arange(0, n), (n, 1))

    ab, bb, X_sb, X_tb = nx.from_numpy(a, b, X_s, X_t)

    Q, R, g, log = ot.lowrank.lowrank_sinkhorn(X_sb, X_tb, ab, bb, reg=0.1, log=True)
    lazy_plan = log["lazy_plan"]
    P = lazy_plan[:]

    np.testing.assert_allclose(ab, P.sum(1), atol=1e-05)
    np.testing.assert_allclose(bb, P.sum(0), atol=1e-05)



# -*- coding: utf-8 -*-
"""
Efficient 1D solver for the partial optimal transport problem.
"""

# Author: Romain Tavenard <romain.tavenard@univ-rennes2.fr>
#
# License: MIT License

# import compiled emd
from .partial_solvers import (
    partial_wasserstein_lagrange,
    partial_wasserstein,
    partial_wasserstein2,
    entropic_partial_wasserstein,
    gwgrad_partial,
    gwloss_partial,
    partial_gromov_wasserstein,
    partial_gromov_wasserstein2,
    entropic_partial_gromov_wasserstein,
    entropic_partial_gromov_wasserstein2,
    partial_wasserstein_1d,
)

__all__ = [
    "partial_wasserstein_1d",
    "partial_wasserstein_lagrange",
    "partial_wasserstein",
    "partial_wasserstein2",
    "entropic_partial_wasserstein",
    "gwgrad_partial",
    "gwloss_partial",
    "partial_gromov_wasserstein",
    "partial_gromov_wasserstein2",
    "entropic_partial_gromov_wasserstein",
    "entropic_partial_gromov_wasserstein2",
]



# -*- coding: utf-8 -*-
"""
Dimension reduction with OT


.. warning::
    Note that by default the module is not imported in :mod:`ot`. In order to
    use it you need to explicitly import :mod:`ot.dr`

"""

# Author: Remi Flamary <remi.flamary@unice.fr>
#         Minhui Huang <mhhuang@ucdavis.edu>
#         Jakub Zadrozny <jakub.r.zadrozny@gmail.com>
#         Antoine Collas <antoine.collas@inria.fr>
#
# License: MIT License

from scipy import linalg

try:
    import autograd.numpy as np
    from sklearn.decomposition import PCA

    import pymanopt
    import pymanopt.manifolds
    import pymanopt.optimizers
except ImportError:
    raise ImportError(
        "Missing dependency for ot.dr. Requires autograd, pymanopt, scikit-learn. You can install with install with 'pip install POT[dr]', or 'conda install autograd pymanopt scikit-learn'"
    )

from .bregman import sinkhorn as sinkhorn_bregman
from .utils import dist as dist_utils, check_random_state


def dist(x1, x2):
    r"""Compute squared euclidean distance between samples (autograd)"""
    x1p2 = np.sum(np.square(x1), 1)
    x2p2 = np.sum(np.square(x2), 1)
    return x1p2.reshape((-1, 1)) + x2p2.reshape((1, -1)) - 2 * np.dot(x1, x2.T)


def sinkhorn(w1, w2, M, reg, k):
    r"""Sinkhorn algorithm with fixed number of iteration (autograd)"""
    K = np.exp(-M / reg)
    ui = np.ones((M.shape[0],))
    vi = np.ones((M.shape[1],))
    for i in range(k):
        vi = w2 / (np.dot(K.T, ui) + 1e-50)
        ui = w1 / (np.dot(K, vi) + 1e-50)
    G = ui.reshape((M.shape[0], 1)) * K * vi.reshape((1, M.shape[1]))
    return G


def logsumexp(M, axis):
    r"""Log-sum-exp reduction compatible with autograd (no numpy implementation)"""
    amax = np.amax(M, axis=axis, keepdims=True)
    return np.log(np.sum(np.exp(M - amax), axis=axis)) + np.squeeze(amax, axis=axis)


def sinkhorn_log(w1, w2, M, reg, k):
    r"""Sinkhorn algorithm in log-domain with fixed number of iteration (autograd)"""
    Mr = -M / reg
    ui = np.zeros((M.shape[0],))
    vi = np.zeros((M.shape[1],))
    log_w1 = np.log(w1)
    log_w2 = np.log(w2)
    for i in range(k):
        vi = log_w2 - logsumexp(Mr + ui[:, None], 0)
        ui = log_w1 - logsumexp(Mr + vi[None, :], 1)
    G = np.exp(ui[:, None] + Mr + vi[None, :])
    return G


def split_classes(X, y):
    r"""split samples in :math:`\mathbf{X}` by classes in :math:`\mathbf{y}`"""
    lstsclass = np.unique(y)
    return [X[y == i, :].astype(np.float32) for i in lstsclass]


def fda(X, y, p=2, reg=1e-16):
    r"""Fisher Discriminant Analysis

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Training samples.
    y : ndarray, shape (n,)
        Labels for training samples.
    p : int, optional
        Size of dimensionality reduction.
    reg : float, optional
        Regularization term >0 (ridge regularization)

    Returns
    -------
    P : ndarray, shape (d, p)
        Optimal transportation matrix for the given parameters
    proj : callable
        projection function including mean centering
    """

    mx = np.mean(X)
    X -= mx.reshape((1, -1))

    # data split between classes
    d = X.shape[1]
    xc = split_classes(X, y)
    nc = len(xc)

    p = min(nc - 1, p)

    Cw = 0
    for x in xc:
        Cw += np.cov(x, rowvar=False)
    Cw /= nc

    mxc = np.zeros((d, nc))

    for i in range(nc):
        mxc[:, i] = np.mean(xc[i])

    mx0 = np.mean(mxc, 1)
    Cb = 0
    for i in range(nc):
        Cb += (mxc[:, i] - mx0).reshape((-1, 1)) * (mxc[:, i] - mx0).reshape((1, -1))

    w, V = linalg.eig(Cb, Cw + reg * np.eye(d))

    idx = np.argsort(w.real)

    Popt = V[:, idx[-p:]]

    def proj(X):
        return (X - mx.reshape((1, -1))).dot(Popt)

    return Popt, proj


def wda(
    X,
    y,
    p=2,
    reg=1,
    k=10,
    solver=None,
    sinkhorn_method="sinkhorn",
    maxiter=100,
    verbose=0,
    P0=None,
    normalize=False,
):
    r"""
    Wasserstein Discriminant Analysis :ref:`[11] <references-wda>`

    The function solves the following optimization problem:

    .. math::
        \mathbf{P} = \mathop{\arg \min}_\mathbf{P} \quad
        \frac{\sum\limits_i W(P \mathbf{X}^i, P \mathbf{X}^i)}{\sum\limits_{i, j \neq i} W(P \mathbf{X}^i, P \mathbf{X}^j)}

    where :

    - :math:`P` is a linear projection operator in the Stiefel(`p`, `d`) manifold
    - :math:`W` is entropic regularized Wasserstein distances
    - :math:`\mathbf{X}^i` are samples in the dataset corresponding to class i

    **Choosing a Sinkhorn solver**

    By default and when using a regularization parameter that is not too small
    the default sinkhorn solver should be enough. If you need to use a small
    regularization to get sparse cost matrices, you should use the
    :py:func:`ot.dr.sinkhorn_log` solver that will avoid numerical
    errors, but can be slow in practice.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Training samples.
    y : ndarray, shape (n,)
        Labels for training samples.
    p : int, optional
        Size of dimensionality reduction.
    reg : float, optional
        Regularization term >0 (entropic regularization)
    solver : None | str, optional
        None for steepest descent or 'TrustRegions' for trust regions algorithm
        else should be a pymanopt.solvers
    sinkhorn_method : str
        method used for the Sinkhorn solver, either 'sinkhorn' or 'sinkhorn_log'
    P0 : ndarray, shape (d, p)
        Initial starting point for projection.
    normalize : bool, optional
        Normalize the Wasserstaiun distance by the average distance on P0 (default : False)
    verbose : int, optional
        Print information along iterations.

    Returns
    -------
    P : ndarray, shape (d, p)
        Optimal transportation matrix for the given parameters
    proj : callable
        Projection function including mean centering.


    .. _references-wda:
    References
    ----------
    .. [11] Flamary, R., Cuturi, M., Courty, N., & Rakotomamonjy, A. (2016).
        Wasserstein Discriminant Analysis. arXiv preprint arXiv:1608.08063.
    """  # noqa

    if sinkhorn_method.lower() == "sinkhorn":
        sinkhorn_solver = sinkhorn
    elif sinkhorn_method.lower() == "sinkhorn_log":
        sinkhorn_solver = sinkhorn_log
    else:
        raise ValueError("Unknown Sinkhorn method '%s'." % sinkhorn_method)

    mx = np.mean(X)
    X -= mx.reshape((1, -1))

    # data split between classes
    d = X.shape[1]
    xc = split_classes(X, y)
    # compute uniform weighs
    wc = [np.ones((x.shape[0]), dtype=np.float32) / x.shape[0] for x in xc]

    # pre-compute reg_c,c'
    if P0 is not None and normalize:
        regmean = np.zeros((len(xc), len(xc)))
        for i, xi in enumerate(xc):
            xi = np.dot(xi, P0)
            for j, xj in enumerate(xc[i:]):
                xj = np.dot(xj, P0)
                M = dist(xi, xj)
                regmean[i, j] = np.sum(M) / (len(xi) * len(xj))
    else:
        regmean = np.ones((len(xc), len(xc)))

    manifold = pymanopt.manifolds.Stiefel(d, p)

    @pymanopt.function.autograd(manifold)
    def cost(P):
        # wda loss
        loss_b = 0
        loss_w = 0

        for i, xi in enumerate(xc):
            xi = np.dot(xi, P)
            for j, xj in enumerate(xc[i:]):
                xj = np.dot(xj, P)
                M = dist(xi, xj)
                G = sinkhorn_solver(wc[i], wc[j + i], M, reg * regmean[i, j], k)
                if j == 0:
                    loss_w += np.sum(G * M)
                else:
                    loss_b += np.sum(G * M)

        # loss inversed because minimization
        return loss_w / loss_b

    # declare manifold and problem

    problem = pymanopt.Problem(manifold=manifold, cost=cost)

    # declare solver and solve
    if solver is None:
        solver = pymanopt.optimizers.SteepestDescent(
            max_iterations=maxiter, log_verbosity=verbose
        )
    elif solver in ["tr", "TrustRegions"]:
        solver = pymanopt.optimizers.TrustRegions(
            max_iterations=maxiter, log_verbosity=verbose
        )

    Popt = solver.run(problem, initial_point=P0)

    def proj(X):
        return (X - mx.reshape((1, -1))).dot(Popt.point)

    return Popt.point, proj


def projection_robust_wasserstein(
    X,
    Y,
    a,
    b,
    tau,
    U0=None,
    reg=0.1,
    k=2,
    stopThr=1e-3,
    maxiter=100,
    verbose=0,
    random_state=None,
):
    r"""
    Projection Robust Wasserstein Distance :ref:`[32] <references-projection-robust-wasserstein>`

    The function solves the following optimization problem:

    .. math::
        \max_{U \in St(d, k)} \ \min_{\pi \in \Pi(\mu,\nu)} \quad \sum_{i,j} \pi_{i,j}
        \|U^T(\mathbf{x}_i - \mathbf{y}_j)\|^2 - \mathrm{reg} \cdot H(\pi)

    - :math:`U` is a linear projection operator in the Stiefel(`d`, `k`) manifold
    - :math:`H(\pi)` is entropy regularizer
    - :math:`\mathbf{x}_i`, :math:`\mathbf{y}_j` are samples of measures :math:`\mu` and :math:`\nu` respectively

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Samples from measure :math:`\mu`
    Y : ndarray, shape (n, d)
        Samples from measure :math:`\nu`
    a : ndarray, shape (n, )
        weights for measure :math:`\mu`
    b : ndarray, shape (n, )
        weights for measure :math:`\nu`
    tau : float
        stepsize for Riemannian Gradient Descent
    U0 : ndarray, shape (d, p)
        Initial starting point for projection.
    reg : float, optional
        Regularization term >0 (entropic regularization)
    k : int
        Subspace dimension
    stopThr : float, optional
        Stop threshold on error (>0)
    verbose : int, optional
        Print information along iterations.
    random_state : int, RandomState instance or None, default=None
        Determines random number generation for initial value of projection
        operator when U0 is not given.

    Returns
    -------
    pi : ndarray, shape (n, n)
        Optimal transportation matrix for the given parameters
    U : ndarray, shape (d, k)
        Projection operator.


    .. _references-projection-robust-wasserstein:
    References
    ----------
    .. [32] Huang, M. , Ma S. & Lai L. (2021).
        A Riemannian Block Coordinate Descent Method for Computing
        the Projection Robust Wasserstein Distance, ICML.
    """  # noqa

    # initialization
    n, d = X.shape
    m, d = Y.shape
    a = np.asarray(a, dtype=np.float64)
    b = np.asarray(b, dtype=np.float64)
    u = np.ones(n) / n
    v = np.ones(m) / m
    ones = np.ones((n, m))

    assert d > k

    if U0 is None:
        rng = check_random_state(random_state)
        U = rng.randn(d, k)
        U, _ = np.linalg.qr(U)
    else:
        U = U0

    def Vpi(X, Y, a, b, pi):
        # Return the second order matrix of the displacements: sum_ij { (pi)_ij (X_i-Y_j)(X_i-Y_j)^T }.
        A = X.T.dot(pi).dot(Y)
        return (
            X.T.dot(np.diag(a)).dot(X)
            + Y.T.dot(np.diag(np.sum(pi, 0))).dot(Y)
            - A
            - A.T
        )

    err = 1
    iter = 0

    while err > stopThr and iter < maxiter:
        # Projected cost matrix
        UUT = U.dot(U.T)
        M = (
            np.diag(np.diag(X.dot(UUT.dot(X.T)))).dot(ones)
            + ones.dot(np.diag(np.diag(Y.dot(UUT.dot(Y.T)))))
            - 2 * X.dot(UUT.dot(Y.T))
        )

        A = np.empty(M.shape, dtype=M.dtype)
        np.divide(M, -reg, out=A)
        np.exp(A, out=A)

        # Sinkhorn update
        Ap = (1 / a).reshape(-1, 1) * A
        AtransposeU = np.dot(A.T, u)
        v = np.divide(b, AtransposeU)
        u = 1.0 / np.dot(Ap, v)
        pi = u.reshape((-1, 1)) * A * v.reshape((1, -1))

        V = Vpi(X, Y, a, b, pi)

        # Riemannian gradient descent
        G = 2 / reg * V.dot(U)
        GTU = G.T.dot(U)
        xi = G - U.dot(GTU + GTU.T) / 2  # Riemannian gradient
        U, _ = np.linalg.qr(U + tau * xi)  # Retraction by QR decomposition

        grad_norm = np.linalg.norm(xi)
        err = max(reg * grad_norm, np.linalg.norm(np.sum(pi, 0) - b, 1))

        f_val = np.trace(U.T.dot(V.dot(U)))
        if verbose:
            print("RBCD Iteration: ", iter, " error", err, "\t fval: ", f_val)

        iter = iter + 1

    return pi, U


def ewca(
    X,
    U0=None,
    reg=1,
    k=2,
    method="BCD",
    sinkhorn_method="sinkhorn",
    stopThr=1e-6,
    maxiter=100,
    maxiter_sink=1000,
    maxiter_MM=10,
    verbose=0,
):
    r"""
    Entropic Wasserstein Component Analysis :ref:`[52] <references-entropic-wasserstein-component_analysis>`.

    The function solves the following optimization problem:

    .. math::
        \mathbf{U} = \mathop{\arg \min}_\mathbf{U} \quad
        W(\mathbf{X}, \mathbf{U}\mathbf{U}^T \mathbf{X})

    where :

    - :math:`\mathbf{U}` is a matrix in the Stiefel(`p`, `d`) manifold
    - :math:`W` is entropic regularized Wasserstein distances
    - :math:`\mathbf{X}` are samples

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Samples from measure :math:`\mu`.
    U0 : ndarray, shape (d, k), optional
        Initial starting point for projection.
    reg : float, optional
        Regularization term >0 (entropic regularization).
    k : int, optional
        Subspace dimension.
    method : str, optional
        Eather 'BCD' or 'MM' (Block Coordinate Descent or Majorization-Minimization).
        Prefer MM when d is large.
    sinkhorn_method : str
        Method used for the Sinkhorn solver, see :ref:`ot.bregman.sinkhorn` for more details.
    stopThr : float, optional
        Stop threshold on error (>0).
    maxiter : int, optional
        Maximum number of iterations of the BCD/MM.
    maxiter_sink : int, optional
        Maximum number of iterations of the Sinkhorn solver.
    maxiter_MM : int, optional
        Maximum number of iterations of the MM (only used when method='MM').
    verbose : int, optional
        Print information along iterations.

    Returns
    -------
    pi : ndarray, shape (n, n)
        Optimal transportation matrix for the given parameters.
    U : ndarray, shape (d, k)
        Matrix Stiefel manifold.


    .. _references-entropic-wasserstein-component_analysis:
    References
    ----------
    .. [52] Collas, A., Vayer, T., Flamary, F., & Breloy, A. (2023).
        Entropic Wasserstein Component Analysis.
    """  # noqa
    n, d = X.shape
    X = X - X.mean(0)

    if U0 is None:
        pca_fitted = PCA(n_components=k).fit(X)
        U = pca_fitted.components_.T
        if method == "MM":
            lambda_scm = pca_fitted.explained_variance_[0]
    else:
        U = U0

    # marginals
    u0 = (1.0 / n) * np.ones(n)

    # print iterations
    if verbose > 0:
        print(
            "{:4s}|{:13s}|{:12s}|{:12s}".format("It.", "Loss", "Crit.", "Thres.")
            + "\n"
            + "-" * 40
        )

    def compute_loss(M, pi, reg):
        return np.sum(M * pi) + reg * np.sum(pi * (np.log(pi) - 1))

    def grassmann_distance(U1, U2):
        proj = U1.T @ U2
        _, s, _ = np.linalg.svd(proj)
        s[s > 1] = 1
        s = np.arccos(s)
        return np.linalg.norm(s)

    # loop
    it = 0
    crit = np.inf
    sinkhorn_warmstart = None

    while (it < maxiter) and (crit > stopThr):
        U_old = U

        # Solve transport
        M = dist_utils(X, (X @ U) @ U.T)
        pi, log_sinkhorn = sinkhorn_bregman(
            u0,
            u0,
            M,
            reg,
            numItermax=maxiter_sink,
            method=sinkhorn_method,
            warmstart=sinkhorn_warmstart,
            warn=False,
            log=True,
        )
        key_warmstart = "warmstart"
        if key_warmstart in log_sinkhorn:
            sinkhorn_warmstart = log_sinkhorn[key_warmstart]
        if (pi >= 1e-300).all():
            loss = compute_loss(M, pi, reg)
        else:
            loss = np.inf

        # Solve PCA
        pi_sym = (pi + pi.T) / 2

        if method == "BCD":
            # block coordinate descent
            S = X.T @ (2 * pi_sym - (1.0 / n) * np.eye(n)) @ X
            _, U = np.linalg.eigh(S)
            U = U[:, ::-1][:, :k]

        elif method == "MM":
            # majorization-minimization
            eig, _ = np.linalg.eigh(pi_sym)
            lambda_pi = eig[0]

            for _ in range(maxiter_MM):
                X_proj = X @ U
                X_T_X_proj = X.T @ X_proj

                R = (1 / n) * X_T_X_proj
                alpha = 1 - 2 * n * lambda_pi
                if alpha > 0:
                    R = alpha * (R - lambda_scm * U)
                else:
                    R = alpha * R

                R = R - (2 * X.T @ (pi_sym @ X_proj)) + (2 * lambda_pi * X_T_X_proj)
                U, _ = np.linalg.qr(R)

        else:
            raise ValueError(f"Unknown method '{method}', use 'BCD' or 'MM'.")

        # stop or not
        it += 1
        crit = grassmann_distance(U_old, U)

        # print
        if verbose > 0:
            print("{:4d}|{:8e}|{:8e}|{:8e}".format(it, loss, crit, stopThr))

    return pi, U
